#ifndef TREE_H
#define TREE_H

#include <containers/stack.h>
#include <containers/vec_.h>
#include <sys/region.h>
#include <types.h>

typedef struct TreeNode {
	void const*			data;
	struct TreeNode* 	parent;
	Vec					children;
} TreeNode;

static inline void attachChild_TreeNode( TreeNode* self, TreeNode child );
static inline void attachChildren_TreeNode( TreeNode* self, TreeNode children[], size_t numChildren ) {
	for( size_t i=0; i<numChildren; ++i ) {
		attachChild_TreeNode( self, children[i] );
	}
}

static inline void init_TreeNode( TreeNode* self, void const* data, void (*destroy_func)(void*), TreeNode children[], size_t numChildren ) {
	self->parent = NULL;
	self->data = data;
	self->children = create_Vec( sizeof(TreeNode), Default_RegionAlignment, destroy_func );

	attachChildren_TreeNode( self, children, numChildren );
}

static inline TreeNode create_TreeNode( void const* data, void (*destroy_func)(void*), TreeNode children[], size_t numChildren ) {
	TreeNode retval;
	init_TreeNode( &retval, data, destroy_func, children, numChildren );

	return retval;
}

static inline void destroy_TreeNode( TreeNode* self ) {
	destroy_Vec( &self->children );
}

static inline TreeNode* new_TreeNode( void const* data, void (*destroy_func)(void*) ) {
	TreeNode* node = malloc( sizeof(*node) );
	assert( NULL != node );
	init_TreeNode( node, data, destroy_func, NULL, 0 );

	return node;
}

static inline void delete_TreeNode( TreeNode* self ) {
	destroy_TreeNode( self );
	free( self );
}

static inline void destroy_TreeNode_( void* self ) {
	destroy_TreeNode( (TreeNode*)self );
}

static inline TreeNode clone_TreeNode( TreeNode const* self, void* (*clone_func)(void const*) ) {
	abort();
	TreeNode cloned;
	init_TreeNode( &cloned,
		clone_func ? clone_func( self->data ) : self->data,
		self->children.base.destroy_func,
		NULL,
		0
	);

	return cloned;
}

static inline void move_TreeNode( TreeNode* self ) {
	VecIterator iter = createIterator_Vec( &self->children, (IteratorRange) { .type = All_IteratorRange } );
	for( TreeNode* child; (child = next_VecIterator( &iter )); ) {
		child->parent = self;
		move_TreeNode( child );
	}
	destroy_VecIterator( &iter );
}

static inline void attachChild_TreeNode( TreeNode* self, TreeNode child ) {
	child.parent = self;
	pushBack_Vec( &self->children, &child );
}

static inline size_t getNumChildren_TreeNode( TreeNode const* self ) {
	return getLength_Vec( &self->children );
}

static inline TreeNode* getParent_TreeNode( TreeNode const* self ) {
	return self->parent;
}

static inline TreeNode* getChild_TreeNode( TreeNode const* self, uint_t i ) {
	return (TreeNode*)getConstElement_Vec( &self->children, i );
}

static inline void const* getData_TreeNode( TreeNode const* self ) {
	return self->data;
}

typedef struct Tree {
	TreeNode	root;
	void*		(*clone_func)(void const*);
	void		(*destroy_func)(void*);
} Tree;

static inline void init_Tree( Tree* self, TreeNode root, void* (*clone_func)(void const*), void (*destroy_func)(void*) ) {
	self->root = root;
	move_TreeNode( &self->root );
	self->clone_func = clone_func;
	self->destroy_func = destroy_func;
}

/*
 * move required!
 */
static inline Tree create_Tree( TreeNode root, void* (*clone_func)(void const*), void (*destroy_func)(void*) ) {
	Tree retval;
	init_Tree( &retval, root, clone_func, destroy_func );

	return retval;
}

void destroy_Tree( Tree* self );
Tree clone_Tree( Tree const* self );

static inline TreeNode const* getRootConst_Tree( Tree const* self ) {
	return &self->root;
}

static inline TreeNode* getRoot_Tree( Tree* self ) {
	return &self->root;
}

typedef struct {
	Stack			stack;		// Stack(size_t)
	Stack			parent;		// Stack(TreeNode*)
	TreeNode const*	node;
	TreeNode const*	next;
	size_t			i;
} TreeNodeIterator;

static inline void reset_TreeNodeIterator( TreeNodeIterator* self, TreeNode const* node ) {
	self->node = node;
	self->next = self->node;
	self->i = 0;
}

static inline void init_TreeNodeIterator( TreeNodeIterator* self, TreeNode const* node ) {
	init_Stack( &self->stack, sizeof(self->i), NULL );
	init_Stack( &self->parent, sizeof(TreeNode*), NULL );
	reset_TreeNodeIterator( self, node );
}

static inline TreeNodeIterator create_TreeNodeIterator( TreeNode const* node ) {
	TreeNodeIterator retval;
	init_TreeNodeIterator( &retval, node );

	return retval;
}

static inline void destroy_TreeNodeIterator( TreeNodeIterator* self ) {
	destroy_Stack( &self->parent );
	destroy_Stack( &self->stack );
}

TreeNode const* inc_TreeNodeIterator( TreeNodeIterator* self );

#endif	/* TREE_H */

