#ifndef VEC__H
#define VEC__H

#include <assert.h>
#include <containers/iterator.h>
#include <malloc.h>
#include <string.h>
#include <types.h>

typedef struct {
	Container	base;
	size_t		num_el;
	size_t 		capacity;
	void*		data_unaligned;
	void*		data;
	void*		end;
} Vec;

static inline size_t align_size_vec( size_t const el_size, size_t const num_el, uint_t const alignment ) {
	return (el_size * num_el + ((1 << alignment) - 1)) & ~((1 << alignment) - 1);
}

void init_Vec( Vec* self, size_t const el_size, size_t const alignment, void (*destroy_func)(void*) );
void destroy_Vec( Vec* self );
Vec clone_Vec( Vec const* self, void* (*clone_func)(void const*) );
Vec* new_Vec( size_t const el_size, size_t const alignment, void (*destroy_func)(void*) );
void append_Vec( Vec* self, Vec const* other );

static inline void resize_Vec( Vec* self, size_t const newLength ) {
	if( self->capacity >= newLength ) {
		return;
	}

	self->data_unaligned =
		realloc(
			self->data_unaligned,
			align_size_vec( self->base.el_size, newLength, self->base.alignment )
		);
	assert( NULL != self->data_unaligned );

	self->data = ALIGN_ADR( self->data_unaligned, self->base.alignment );
	self->end = (void*)&((char*)self->data)[ self->base.el_size * self->num_el ];
	self->capacity = newLength;
}

static inline void pushBack_Vec( Vec* self, void const* element ) {
	if( self->num_el >= self->capacity ) {
		resize_Vec( self, self->capacity << 1 );
	}

	memcpy( self->end, element, self->base.el_size );
	self->end = (char*)self->end + self->base.el_size;
	self->num_el++;
}

static inline Vec create_Vec( size_t const el_size, size_t const alignment, void (*destroy_func)(void*) ) {
	Vec vec;
	init_Vec( &vec, el_size, alignment, destroy_func );

	return vec;
}

static inline Vec createFromArray_Vec( void const* from, size_t const el_size, size_t const alignment, size_t const num_el ) {
	Vec retval = create_Vec( el_size, alignment, NULL );
	for( size_t i=0; i<num_el; ++i ) {
		pushBack_Vec( &retval, &((char const*)from)[i*el_size] );
	}

	return retval;
}

static inline void delete_Vec( Vec* self ) {
	destroy_Vec( self );
	free( self );
}

static inline size_t getCapacity_Vec( Vec const* self ) {
	return self->capacity;
}

static inline void* getElement_Vec( Vec* self, size_t const i ) {
	return &((char*)self->data)[i * self->base.el_size];
}

static inline void const* getConstElement_Vec( Vec const* self, size_t const i ) {
	return &((char const*)self->data)[i * self->base.el_size];
}

static inline void* getHead_Vec( Vec* self ) {
	return self->data;
}

static inline void const* getHeadConst_Vec( Vec const* self ) {
	return self->data;
}

static inline void const* getTailConst_Vec( Vec const* self ) {
	return getConstElement_Vec( self, self->num_el );
}

static inline void* getTail_Vec( Vec* self ) {
	return getElement_Vec( self, self->num_el );
}

static inline void const* getLastConst_Vec( Vec const* self ) {
	assert( self->num_el > 0 );
	return (char const*)getTailConst_Vec( self ) - self->base.el_size;
}

static inline void* getLast_Vec( Vec* self ) {
	assert( self->num_el > 0 );
	return (char*)getTail_Vec( self ) - self->base.el_size;
}

static inline size_t getElementSize_Vec( Vec const* self ) {
	return self->base.el_size;
}

static inline size_t getLength_Vec( Vec const* self ) {
	return self->num_el;
}

static inline size_t getSizeInBytes_Vec( Vec const* self ) {
	return getElementSize_Vec( self ) * getLength_Vec( self );
}

static inline void reset_Vec( Vec* self ) {
	self->num_el = 0;
	self->end = self->data;
}

typedef struct {
	Iterator	base;
	size_t		num_elements;
	size_t		current_idx;
	void*		current_element;
} VecIterator;

static inline void* next_VecIterator( VecIterator* self ) {
	if( self->current_idx < self->num_elements ) {
		void* ret = self->current_element;

		self->current_element = (void*)((uintptr_t)self->current_element + getElementSize_Vec( (Vec const*)self->base.container ));
		self->current_idx++;

		return ret;
	} else {
		return NULL;
	}
}

static inline void* next_VecIterator_( struct Iterator* self ) {
	return next_VecIterator( (VecIterator*)self );
}

static inline void init_VecIterator( VecIterator* self, Vec* vec, IteratorRange const range ) {
	self->base = create_Iterator( &vec->base, next_VecIterator_ );
	self->current_idx = 0;
	if( All_IteratorRange == range.type ) {
		self->current_element = getHead_Vec( vec );
		self->num_elements = getLength_Vec( vec );
	} else if( To_IteratorRange == range.type ) {
		self->current_element = getHead_Vec( vec );
		self->num_elements = range.to;
	} else if( From_IteratorRange == range.type ) {
		self->current_element = getElement_Vec( vec, range.from );
		self->num_elements = getLength_Vec( vec ) - range.from;
	} else if( FromTo_IteratorRange == range.type ) {
		self->current_element = getElement_Vec( vec, range.from );
		self->num_elements = range.to - range.from;
	}
}

static inline VecIterator createIterator_Vec( Vec* self, IteratorRange const range ) {
	VecIterator it;
	init_VecIterator( &it, self, range );

	return it;
}

static inline void destroy_VecIterator( VecIterator* self ) {
	if( self ) {}
}

typedef struct {
	IteratorConst	base;
	size_t			num_elements;
	size_t			current_idx;
	void const*		current_element;
} VecConstIterator;

static inline void const* next_VecConstIterator( VecConstIterator* self ) {
	if( self->current_idx < self->num_elements ) {
		void const* ret = self->current_element;

		self->current_element = (void const*)((uintptr_t)self->current_element + getElementSize_Vec( (Vec const*)self->base.container ));
		self->current_idx++;

		return ret;
	} else {
		return NULL;
	}
}

static inline void const* next_VecConstIterator_( struct IteratorConst* self ) {
	return next_VecConstIterator( (VecConstIterator*)self );
}

static inline void init_VecConstIterator( VecConstIterator* self, Vec const* vec, IteratorRange const range ) {
	self->base = create_IteratorConst( &vec->base, next_VecConstIterator_ );
	self->current_idx = 0;
	if( All_IteratorRange == range.type ) {
		self->current_element = getHeadConst_Vec( vec );
		self->num_elements = getLength_Vec( vec );
	} else if( To_IteratorRange == range.type ) {
		self->current_element = getHeadConst_Vec( vec );
		self->num_elements = range.to;
	} else if( From_IteratorRange == range.type ) {
		self->current_element = getConstElement_Vec( vec, range.from );
		self->num_elements = getLength_Vec( vec ) - range.from;
	} else if( FromTo_IteratorRange == range.type ) {
		self->current_element = getConstElement_Vec( vec, range.from );
		self->num_elements = range.to - range.from;
	}
}

static inline VecConstIterator createConstIterator_Vec( Vec const* self, IteratorRange const range ) {
	VecConstIterator it;
	init_VecConstIterator( &it, self, range );

	return it;
}

static inline void destroy_VecConstIterator( VecConstIterator* self ) {
	if( self ) {}
}

#endif	/* VEC__H */
