#ifndef SIGNAL_H
#define SIGNAL_H

#include <abort.h>
#include <assert.h>
#include <types.h>

typedef struct {
	float			(*sample_func)( float, void const* );
	void const*		externalArg;
} Signal;

static inline void init_Signal( Signal* self, float (*sample_func)( float, void const* ), void const* externalArg ) {
	*self = (Signal) {
		.sample_func = sample_func,
		.externalArg = externalArg
	};
}

static inline Signal create_Signal( float (*sample_func)( float, void const* ), void const* externalArg ) {
	Signal signal;
	init_Signal( &signal, sample_func, externalArg );

	return signal;
}

static inline void destroy_Signal( Signal* self ) {
	if( self ) {}
}

typedef enum {
	None_SamplerWrapping,
	ClampEdge_SamplerWrapping,
	Wrap_SamplerWrapping
} SamplerWrapping;

typedef struct {
	SamplerWrapping	wrap;
	uint_t			frequency;
	float			argScale;
} Sampler;

static inline Sampler create_Sampler( SamplerWrapping const wrap, uint_t const frequency, float const argScale ) {
	return (Sampler) {
		.wrap = wrap,
		.frequency = frequency,
		.argScale = argScale
	};
}

static inline void destroy_Sampler( Sampler const* self ) {
	if( self ) {}
}


static inline uint_t getFrequency_Sampler( Sampler const* self ) {
	return self->frequency;
}

static inline float getPeriod_Sampler( Sampler const* self ) {
	return 1.0f / self->frequency;
}

static inline float getArgScale_Sampler( Sampler const* self ) {
	return self->argScale;
}

static inline float discretizeParameter_Sampler( Sampler const* self, float const t ) {
	return (int)(t * getFrequency_Sampler( self )) / (float)getFrequency_Sampler( self );
}

static inline float wrap_Sampler( Sampler const* self, float const t ) {
	switch( self->wrap ) {
		case None_SamplerWrapping:
			return t;
		case ClampEdge_SamplerWrapping:
			return t > 1.0f ? 1.0f :
				t < 0.0f ? 0.0f:
				t;
		case Wrap_SamplerWrapping:
			return t - (int)t;
		default:
			abort();
			return t;
	}
}

typedef struct {
	Sampler	base;
} SamplerPoint;

static inline SamplerPoint create_SamplerPoint( Sampler base ) {
	return (SamplerPoint) {
		.base = base
	};
}

static inline void destroy_SamplerPoint( SamplerPoint const* self ) {
	destroy_Sampler( &self->base );
}

static inline float sample_SamplerPoint( SamplerPoint const* self, Signal const* signal, float t ) {
	t = wrap_Sampler( &self->base, discretizeParameter_Sampler( &self->base, t ) );
	return signal->sample_func( t * getArgScale_Sampler( &self->base ), signal->externalArg );
}

typedef struct {
	Sampler base;
} SamplerLinear;

static inline SamplerLinear create_SamplerLinear( Sampler base ) {
	return (SamplerLinear) {
		.base = base
	};
}

static inline void destroy_SamplerLinear( SamplerLinear const* self ) {
	destroy_Sampler( &self->base );
}

static inline float sample_SamplerLinear( SamplerLinear const* self, Signal const* signal, float t ) {
	float const frac_t = wrap_Sampler( &self->base,
		t * getFrequency_Sampler( &self->base ) -
		(int)(t * getFrequency_Sampler( &self->base ))
	);
	t = discretizeParameter_Sampler( &self->base, t );
	float sample0_t = wrap_Sampler( &self->base, t );
	float sample1_t = wrap_Sampler( &self->base, t + getPeriod_Sampler( &self->base ) );

	return
		(1.0f - frac_t) * signal->sample_func( sample0_t * getArgScale_Sampler( &self->base ), signal->externalArg ) +
		frac_t * signal->sample_func( sample1_t * getArgScale_Sampler( &self->base ), signal->externalArg );
}

#endif	/* SIGNAL_H */
