#ifndef ENGINE_H
#define ENGINE_H

#include <containers/tree.h>
#include <containers/vec_.h>
#include <dltrans.h>
#include <math/matrix.h>
#include <math/vector.h>
#include <mempool.h>
#include <mesh.h>
#include <myheader.h>
#include <node/billboard.h>
#include <node/camera.h>
#include <node/flare.h>
#include <node/lux.h>
#include <node/skybox.h>
#include <node/xformer.h>
#include <shader.h>
#include <sys/dlist.h>
#include <sys/nativedl.h>
#include <sys/surface.h>
#include <viewport.h>

typedef enum {
	Raw_ObjectType,
	Scene_ObjectType,
	StaticMesh_ObjectType,
	Mesh_ObjectType,
	Camera_ObjectType,
	Billboard_ObjectType,
	Lux_ObjectType,
	Flare_ObjectType,
	Skybox_ObjectType
} ObjectType;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
} Raw_Xformer;

typedef struct {
	ObjectType		tag;
	Xformer*		xf;
	StaticMesh*		staticMesh;
} StaticMesh_Xformer;

typedef struct {
	ObjectType			tag;
	Xformer*			xf;
	Mesh*				mesh;
	VertexShader const*	vertexShader;
} Mesh_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Billboard*	billboard;
} Billboard_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Camera*		camera;
} Camera_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Lux*		lux;
} Lux_Xformer;

typedef struct {
	StaticMesh_Xformer	staticMeshXf;
	float				scale;
} Pair_StaticMesh_Xformer_Scale;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Vec*		meshes;	// &Vec(Pair_StaticMesh_Xformer_Scale)
} Flare_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Skybox*		skybox;
} Skybox_Xformer;

static inline Mesh_Xformer const* upcast_Mesh_Xformer( Raw_Xformer const* self ) {
	return (Mesh_Xformer const*)self;
}

typedef struct {
	Vec						lights;				// Vec(Lux const*)
	Vec						materials;			// Vec(Material const*)
	Vec						textures;			// Vec(Texture const*)
	Vec						transparentXf;		// Vec(Raw_Xformer*)
	Vec 					opaqueXf;			// Vec(Raw_Xformer*)
	Vec						skyboxesXf;			// Vec(Skybox_Xformer*)
	Vec						cpuMatrixes;		// Vec(Matrix)
	MemPool					xformersPool;		// MemPool(Xformer)
	MemPool					staticMeshesPool;	// MemPool(StaticMesh)
	MemPool					meshesPool;			// MemPool(Mesh)
	MemPool					billboardsPool;		// MemPool(Billboard)
	MemPool					camerasPool;		// MemPool(Camera)
	MemPool					luxesPool;			// MemPool(Lux)
	MemPool					materialsPool;		// MemPool(Material)
	Viewport 				viewport;
	Camera_Xformer const*	cameraXf;
	Camera_Xformer			camera2dXf;
	Surface const*			renderTarget;
	Surface const*			zBuffer;
	Matrix					projMtxF;
	DList					dl;
	DLTrans					dltrans;
} Engine;

#include "task.h"

void init_Engine( Engine* self );

static inline Engine create_Engine( void ) {
	Engine retval;
	init_Engine( &retval );

	return retval;
}

static inline void setViewport_Engine( Engine* self, Viewport const viewport ) {
	self->viewport = viewport;
}

static inline void setCamera_Engine( Engine* self, Camera_Xformer const* cameraXf ) {
	self->cameraXf = cameraXf;
}

static inline void setRenderTarget_Engine( Engine* self, Surface const* renderTarget ) {
	self->renderTarget = renderTarget;
}

static inline void setZBuffer_Engine( Engine* self, Surface const* zBuffer ) {
	self->zBuffer = zBuffer;
}

static inline Viewport const* getViewport_Engine( Engine const* self ) {
	return &self->viewport;
}

static inline Camera_Xformer const* getCamera_Engine( Engine const* self ) {
	return self->cameraXf;
}

static inline Surface const* getRenderTarget_Engine( Engine const* self ) {
	return self->renderTarget;
}

static inline Surface const* getZBuffer_Engine( Engine const* self ) {
	return self->zBuffer;
}

static inline DList* getDisplayList_Engine( Engine* self ) {
	return &self->dl;
}

static inline size_t getMaterialsNum_Engine( Engine const* self ) {
	return getLength_Vec( &self->materials );
}

static inline size_t getTexturesNum_Engine( Engine const* self ) {
	return getLength_Vec( &self->textures );
}

static inline Texture const* getTextureConstReference_Engine( Engine const* self, size_t textureId ) {
	return *(Texture const**)getConstElement_Vec( &self->textures, textureId );
}

static inline void addLight_Engine( Engine* self, Lux const* light ) {
	pushBack_Vec( &self->lights, &light );
}

static inline size_t addMaterialReference_Engine( Engine* self, Material const* material ) {
	size_t const id = getMaterialsNum_Engine( self );
	pushBack_Vec( &self->materials, &material );

	return id;
}

static inline Material* getMaterialReference_Engine( Engine* self, size_t const materialId ) {
	return *(Material**)getElement_Vec( &self->materials, materialId );
}

static inline size_t addTextureReference_Engine( Engine* self, Texture const* texture ) {
	size_t const id = getTexturesNum_Engine( self );
	pushBack_Vec( &self->textures, &texture );

	return id;
}

static inline void removeLight_Engine( Engine* self, Lux const* light ) {
	if( self ) {}
	if( light ) {}
}

void processTask_Engine( Engine* self, Task const* task );
void sceneTaskF_Engine( Engine* self, Tree const* scene );
void flareTaskF_Engine( Engine* self, Tree const* scene );
void overlayTaskF_Engine( Engine* self, void* arg );
void clearColorbufF_Engine( Engine* self, void* arg );
void clearZbufF_Engine( Engine* self, void* arg );

Xformer*			createXformer_Engine( Engine* self, ProcessFunc_Xformer processor );
void				freeXformer_Engine( Engine* self, Xformer* xf );
Raw_Xformer			createRaw_Engine( Engine* self );
void				freeRaw_Engine( Engine* self, Raw_Xformer* xf );
StaticMesh_Xformer	createStaticMesh_Engine( Engine* self, StaticMesh const* staticMesh );
void				freeStaticMesh_Engine( Engine* self, StaticMesh_Xformer* staticMeshXf );
Mesh_Xformer		createMesh_Engine( Engine* self, Mesh meshFrom, Material const* material, VertexShader const* vertexShader );
void				freeMesh_Engine( Engine* self, Mesh_Xformer* meshXf );
Billboard_Xformer	createBillboard_Engine( Engine* self, StaticMesh const* staticMesh );
Camera_Xformer		createCamera_Engine( Engine* self, CameraType type );
void				freeCamera_Engine( Engine* self, Camera_Xformer* cameraXf );
Lux_Xformer			createLux_Engine( Engine* self );
void				freeLux_Engine( Engine* self, Lux_Xformer* luxXf );
Flare_Xformer		createFlare_Engine( Engine* self, Vec* flare_meshes );
Skybox_Xformer		createSkybox_Engine( Engine* self, Mesh meshFrom, Material const* material );

Material* createMaterial_Engine( Engine* self,
	Vector4f const color,
	TextureCombine const combineMode,
	uint16_t const texture0Id,
	uint16_t const texture1Id
);

#endif	/* ENGINE_H */

