#ifndef LAYER_H
#define LAYER_H

#include <containers/vec_.h>
#include <rectangle.h>
#include <segment.h>
#include <triangle.h>
#include <types.h>
#include <vertex.h>
#include <sys/region.h>

typedef Vertex_pos3f_col4f_n3f_tc2f Vertex_generic;

typedef enum {
	Triangle_PrimitiveType_Layer,
	Segment_PrimitiveType_Layer
} Type_Primitives_Layer;

typedef struct {
	Type_Primitives_Layer	type;
	Vec						primitives;
} Primitives_Layer;

static inline Primitives_Layer create_Primitives_Layer( Type_Primitives_Layer const type ) {
	return (Primitives_Layer) {
		.type = type,
		.primitives = create_Vec(
			(Triangle_PrimitiveType_Layer == type) ? sizeof(Triangle) : sizeof(Segment),
			Default_RegionAlignment,
			NULL
		)
	};
}

static inline void destroy_Primitives_Layer( Primitives_Layer* self ) {
	destroy_Vec( &self->primitives );
}

static inline Primitives_Layer clone_Primitives_Layer( Primitives_Layer const* self ) {
	return (Primitives_Layer) {
		.type = self->type,
		.primitives = clone_Vec( &self->primitives, NULL )
	};
}

static inline Type_Primitives_Layer getType_Primitives_Layer( Primitives_Layer const* primitives ) {
	return primitives->type;
}

typedef struct {
	uint16_t			materialId;
	Vec					verts;			// Vec(Vertex_generic)
	Primitives_Layer	primitives;
	Rectangle			tile;			// temporary hack
} Layer;

static inline void init_Layer( Layer* self, Type_Primitives_Layer const primitiveType ) {
	*self = (Layer) {
		.materialId = 0,
		.verts = create_Vec( sizeof( Vertex_generic ), Default_RegionAlignment, NULL ),
		.primitives = create_Primitives_Layer( primitiveType ),
		.tile = (Rectangle) {
			.p0x = 0, .p0y = 0,
			.p1x = 0, .p1y = 0
		}
	};
}

static inline Layer create_Layer( Type_Primitives_Layer const primitiveType ) {
	Layer retval;
	init_Layer( &retval, primitiveType );

	return retval;
}

static inline void destroy_Layer( Layer* self ) {
	destroy_Primitives_Layer( &self->primitives );
	destroy_Vec( &self->verts );
}

static inline Layer clone_Layer( Layer const* self ) {
	return (Layer) {
		.materialId = self->materialId,
		.verts = clone_Vec( &self->verts, NULL ),
		.primitives = clone_Primitives_Layer( &self->primitives ),
		.tile = self->tile
	};
}

static inline void reset_Layer( Layer* self ) {
	reset_Vec( &self->verts );
	reset_Vec( &self->primitives.primitives );
}

static inline uint16_t getMaterialId_Layer( Layer const* self ) {
	return self->materialId;
}

static inline void setMaterialId_Layer( Layer* self, uint16_t const materialId ) {
	self->materialId = materialId;
}

static inline Type_Primitives_Layer getPrimitiveType_Layer( Layer const* self ) {
	return getType_Primitives_Layer( &self->primitives );
}

static inline Vec const* getVerticesConst_Layer( Layer const* self ) {
	return &self->verts;
}

static inline Vec* getVertices_Layer( Layer* self ) {
	return &self->verts;
}

static inline size_t getVerticesNum_Layer( Layer const* self ) {
	return getLength_Vec( getVerticesConst_Layer( self ) );
}

static inline Primitives_Layer const* getPrimitivesConst_Layer( Layer const* self ) {
	return &self->primitives;
}

static inline Primitives_Layer* getPrimitives_Layer( Layer* self ) {
	return &self->primitives;
}

static inline Vec const* getTrianglesConst_Layer( Layer const* self ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return &getPrimitivesConst_Layer( self )->primitives;
}

static inline Vec* getTriangles_Layer( Layer* self ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return &getPrimitives_Layer( self )->primitives;
}

static inline Triangle* getTriangle_Layer( Layer* self, size_t const triangleIndex ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return (Triangle*)getElement_Vec( getTriangles_Layer( self ), triangleIndex );
}

static inline size_t getTrianglesNum_Layer( Layer const* self ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return getLength_Vec( getTrianglesConst_Layer( self ) );
}
#if 1
static inline Vec const* getSegmentsConst_Layer( Layer const* self ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return &getPrimitivesConst_Layer( self )->primitives;
}

static inline Vec* getSegments_Layer( Layer* self ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return &getPrimitives_Layer( self )->primitives;
}

static inline Segment* getSegment_Layer( Layer* self, size_t const segmentIndex ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return (Segment*)getElement_Vec( getSegments_Layer( self ), segmentIndex );
}

static inline size_t getSegmentsNum_Layer( Layer const* self ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );
	return getLength_Vec( getSegmentsConst_Layer( self ) );
}
#endif
static inline Vertex_generic const* getVertexConst_Layer( Layer const* self, size_t const vertexIndex ) {
	return (Vertex_generic const*)getConstElement_Vec( getVerticesConst_Layer( self ), vertexIndex );
}

static inline Vertex_generic* getVertex_Layer( Layer* self, size_t const vertexIndex ) {
	return (Vertex_generic*)getElement_Vec( getVertices_Layer( self ), vertexIndex );
}

static inline size_t addVertex_Layer( Layer* self, Vertex_generic const vertex ) {
	size_t const indexOfAdded = getVerticesNum_Layer( self );
	pushBack_Vec( getVertices_Layer( self ), &vertex );

	return indexOfAdded;
}

static inline size_t addVertices_Layer( Layer* self, Vec const* vertices ) {
	size_t const indexOfAdded = getVerticesNum_Layer( self );
	append_Vec( getVertices_Layer( self ), vertices );

	return indexOfAdded;
}

static inline size_t addTriangle_Layer( Layer* self, Triangle triangle ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	size_t const indexOfAdded = getTrianglesNum_Layer( self );
	pushBack_Vec( getTriangles_Layer( self ), &triangle );

	return indexOfAdded;
}

static inline size_t addTriangles_Layer( Layer* self, Vec const* triangles ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	size_t const indexOfAdded = getTrianglesNum_Layer( self );
	append_Vec( getTriangles_Layer( self ), triangles );

	return indexOfAdded;
}
#if 1
static inline size_t addSegment_Layer( Layer* self, Segment segment ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	size_t const indexOfAdded = getSegmentsNum_Layer( self );
	pushBack_Vec( getSegments_Layer( self ), &segment );

	return indexOfAdded;
}

static inline size_t addSegments_Layer( Layer* self, Vec const* segments ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	size_t const indexOfAdded = getSegmentsNum_Layer( self );
	append_Vec( getSegments_Layer( self ), segments );

	return indexOfAdded;
}
#endif
static inline VecConstIterator createVertexConstIterator_Layer( Layer const* self ) {
	return createConstIterator_Vec( &self->verts, (IteratorRange) { .type = All_IteratorRange } );
}

static inline VecIterator createVertexIterator_Layer( Layer* self ) {
	return createIterator_Vec( &self->verts, (IteratorRange) { .type = All_IteratorRange } );
}

static inline VecConstIterator createTriangleConstIterator_Layer( Layer const* self ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	return createConstIterator_Vec( &getPrimitivesConst_Layer( self )->primitives, (IteratorRange) { .type = All_IteratorRange } );
}

static inline VecIterator createTriangleIterator_Layer( Layer* self, IteratorRange const range ) {
	assert( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	return createIterator_Vec( &getPrimitives_Layer( self )->primitives, range );
}

// TODO: below is the same as createTriangleIterator - replace both with createPrimitivesIterator
static inline VecConstIterator createSegmentConstIterator_Layer( Layer const* self ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	return createConstIterator_Vec( &getPrimitivesConst_Layer( self )->primitives, (IteratorRange) { .type = All_IteratorRange } );
}

static inline VecIterator createSegmentIterator_Layer( Layer* self, IteratorRange const range ) {
	assert( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( self ) );

	return createIterator_Vec( &getPrimitives_Layer( self )->primitives, range );
}

#endif	/* LAYER_H */
