#ifndef VECTOR_H
#define VECTOR_H

#include <assert.h>
#include <types.h>
#include <math/math.h>

typedef struct {
	float	x;
	float	y;
} Vector2f;

typedef struct {
	uint_t	x;
	uint_t	y;
} Vector2u;

typedef struct {
	float	x;
	float	y;
	float	z;
} Vector3f;

typedef struct {
	int	x;
	int	y;
	int	z;
} Vector3i;

typedef struct {
	uint_t	x;
	uint_t	y;
	uint_t	z;
} Vector3u;

typedef struct {
	float	x;
	float	y;
	float	z;
	float	t;
} Vector4f;

#define VECTOR2F_SET( a,b )	\
{.x = a, .y = b}
#define VECTOR3F_SET( a,b,c ) \
{.x = a, .y = b, .z = c}
#define VECTOR4F_SET( a,b,c,d ) \
{.x = a, .y = b, .z = c, .t = d}
#define VECTOR3F_TEMP( lvalue )  \
(Vector3f) { .x = lvalue.x, .y = lvalue.y, .z = lvalue.z }

static inline Vector2f create_Vector2f( float const x, float const y ) {
	return (Vector2f) {
		.x = x,
		.y = y
	};
}

static inline Vector2u create_Vector2u( uint_t const x, uint_t const y ) {
	return (Vector2u) {
		.x = x,
		.y = y
	};
}

static inline Vector3f create_Vector3f( float const x, float const y, float const z ) {
	return (Vector3f) {
		.x = x,
		.y = y,
		.z = z
	};
}

static inline Vector3f clone_Vector3f( Vector3f const* other ) {
	return *other;
}

static inline Vector3f add_Vector3f( Vector3f const* u, Vector3f const* v ) {
	return (Vector3f) {
		.x = u->x + v->x,
		.y = u->y + v->y,
		.z = u->z + v->z
	};
}

static inline Vector3f saturatedAdd_Vector3f( Vector3f const* u, Vector3f const* v, Vector3f const* bound ) {
	Vector3f result = add_Vector3f( u, v );

	return (Vector3f) {
		.x = fabsf( result.x ) > bound->x ? signum( result.x ) * bound->x : result.x,
		.y = fabsf( result.y ) > bound->y ? signum( result.y ) * bound->y : result.y,
		.z = fabsf( result.z ) > bound->z ? signum( result.z ) * bound->z : result.z
	};
}

static inline Vector3f sub_Vector3f( Vector3f const* u, Vector3f const* v ) {
	return (Vector3f) {
		u->x - v->x,
		u->y - v->y,
		u->z - v->z
	};
}

static inline Vector3f scale_Vector3f( Vector3f const* v, float const scale ) {
	return (Vector3f) {
		v->x * scale,
		v->y * scale,
		v->z * scale
	};
}

static inline Vector3f scaleComponents_Vector3f( Vector3f const* u, Vector3f const* v ) {
	return (Vector3f) {
		u->x * v->x,
		u->y * v->y,
		u->z * v->z
	};
}

static inline float length_Vector3f( Vector3f const* v ) {
	return sqrtf( v->x * v->x + v->y * v->y + v->z * v->z );
}

static inline Vector3f normalize_Vector3f( Vector3f const* v ) {
	float const l = length_Vector3f( v );
	if( l == 0.0f ) {
		return (Vector3f) {
			1.0f,
			1.0f,
			1.0f
		};
	}

	return (Vector3f) {
		v->x / l,
		v->y / l,
		v->z / l
	};
}

Vector3f crossProduct_Vector3f( Vector3f const* u, Vector3f const* v );

static inline bool equals_Vector3f( Vector3f const* u, Vector3f const* v ) {
	return (u->x == v->x) && (u->y == v->y) && (u->z == v->z);
}

static inline bool lessEqual_Vector3f( Vector3f const* u, Vector3f const* v ) {
	return (u->x <= v->x) && (u->y <= v->y) && (u->z <= v->z);
}

static inline Vector3f abs_Vector3f( Vector3f const* self ) {
	return (Vector3f) {
		fabsf( self->x ),
		fabsf( self->y ),
		fabsf( self->z )
	};
}

static inline bool isBounded_Vector3f( Vector3f const* self, Vector3f const* absoluteLimit ) {
	assert(
		absoluteLimit->x >= 0.0f &&
		absoluteLimit->y >= 0.0f &&
		absoluteLimit->z >= 0.0f
	);

	return
		(self->x >= -absoluteLimit->x) && (self->x <= absoluteLimit->x) &&
		(self->y >= -absoluteLimit->y) && (self->y <= absoluteLimit->y) &&
		(self->z >= -absoluteLimit->z) && (self->z <= absoluteLimit->z);
}

static inline Vector3i create_Vector3i( int const x, int const y, int const z ) {
	return (Vector3i) {
		.x = x,
		.y = y,
		.z = z
	};
}

static inline Vector3u create_Vector3u( uint_t const x, uint_t const y, uint_t const z ) {
	return (Vector3u) {
		.x = x,
		.y = y,
		.z = z
	};
}

static inline Vector3i fromVector3f_Vector3i( Vector3f const* other ) {
	return (Vector3i) {
		.x = (other->x),
		.y = (other->y),
		.z = (other->z)
	};
}

static inline Vector4f create_Vector4f( float const x, float const y, float const z, float const t ) {
	return (Vector4f) {
		.x = x,
		.y = y,
		.z = z,
		.t = t
	};
}

#endif	/* VECTOR_H */

