#ifndef MESH_H
#define MESH_H

#include <containers/vec_.h>
#include <layer.h>
#include <material.h>
#include <texture.h>
#include <sys/region.h>

typedef enum {
	MESHATTRIBUTE_RECEIVES_SHADOW = (0x1 << 1),
	MESHATTRIBUTE_CASTS_SHADOW = (0x1 << 0)
} MeshAttribute;

typedef struct {
	Vec				layers;			// Vec(Layer)
	float			opacity;
	MeshAttribute	attributes;
} Mesh;

Mesh clone_Mesh( Mesh const* self );
void reset_Mesh( Mesh* self );

static inline void init_Mesh( Mesh* self ) {
	init_Vec( &self->layers, sizeof( Layer ), Default_RegionAlignment, NULL );
	self->opacity = 1.0f;
	self->attributes = 0;
}

static inline Mesh create_Mesh( void ) {
	Mesh retval;
	init_Mesh( &retval );

	return retval;
}

static inline void destroy_Mesh( Mesh* self ) {
	VecIterator iterator = createIterator_Vec( &self->layers, (IteratorRange) { .type = All_IteratorRange } );
	for( Layer* layer; (layer = (Layer*)next_VecIterator( &iterator )); ) {
		destroy_Layer( layer );
	}
	destroy_VecIterator( &iterator );

	destroy_Vec( &self->layers );
}

static inline void assign_Mesh( Mesh* self, Mesh other ) {
	destroy_Mesh( self );
	*self = other;
}

static inline void setOpacity_Mesh( Mesh* self, float const opacity ) {
	self->opacity = opacity;
}

static inline float getOpacity_Mesh( Mesh const* self ) {
	return self->opacity;
}

static inline bool isOpaque_Mesh( Mesh const* self ) {
	return (1.0f == getOpacity_Mesh( self ));
}

static inline void addLayer_Mesh( Mesh* self, Layer layer ) {
	pushBack_Vec( &self->layers, &layer );
}

static inline Vec const* getLayersConst_Mesh( Mesh const* self ) {
	return &self->layers;
}

static inline Vec* getLayers_Mesh( Mesh* self ) {
	return &self->layers;
}

static inline Layer const* getLayerConst_Mesh( Mesh const* self, size_t const layerIndex ) {
	return (Layer*)getConstElement_Vec( getLayersConst_Mesh( self ), layerIndex );
}

static inline Layer const* getFirstLayerConst_Mesh( Mesh const* self ) {
	return getLayerConst_Mesh( self, 0 );
}

static inline size_t getLayersNum_Mesh( Mesh const* self ) {
	return getLength_Vec( getLayersConst_Mesh( self ) );
}

static inline Layer* getLayer_Mesh( Mesh* self, size_t const layerIndex ) {
	return (Layer*)getElement_Vec( getLayers_Mesh( self ), layerIndex );
}

static inline Layer* getFirstLayer_Mesh( Mesh* self ) {
	return getLayer_Mesh( self, 0 );
}

static inline VecIterator createLayerIterator_Mesh( Mesh* self ) {
	return createIterator_Vec( &self->layers, (IteratorRange) { .type = All_IteratorRange } );
}

static inline VecConstIterator createLayerConstIterator_Mesh( Mesh const* self ) {
	return createConstIterator_Vec( &self->layers, (IteratorRange) { .type = All_IteratorRange } );
}

static inline void setAttributes_Mesh( Mesh* self, MeshAttribute const attributes ) {
	self->attributes |= attributes;
}

static inline MeshAttribute getAttributes_Mesh( Mesh const* self ) {
	return self->attributes;
}

#endif	/* MESH_H */
