#ifndef MYHEADER_H
#define MYHEADER_H

#include <ultra64.h>
#include <malloc.h>
#include <string.h>
#include "math/vector.h"
#include "material.h"
#include "rectangle.h"

typedef enum {
	None_LayerAttributes,
	PreventSeams_LayerAttributes,
	// Deprecated
	LAYERATTRIBUTE_NONE = None_LayerAttributes,
	LAYERATTRIBUTE_PREVENT_SEAMS = PreventSeams_LayerAttributes
} LayerAttributes;

typedef struct {
	uint16_t		material_id;
	Gfx*			gfx;
	uint32_t		gfx_size;
	Vtx*			vts;
	uint32_t		vts_size;
	Rectangle		tile;
	LayerAttributes	attr;
	uint32_t		vts_ofs;
} StaticLayer;

#define STATICLAYER_INIT( mtlid, gfxp, gfxs, vtsp, vtss, t, a ) \
{ .material_id = mtlid, .gfx = gfxp, .gfx_size = gfxs, .vts = vtsp, .vts_size = vtss, .tile = t, .attr = a, .vts_ofs = 0 }

static inline void destroy_StaticLayer( StaticLayer const* self ) {
	free( self->vts );
	free( self->gfx );
}

static inline StaticLayer clone_StaticLayer( StaticLayer const* self ) {
	StaticLayer cloned = *self;

	memcpy(
		(cloned.gfx = malloc( self->gfx_size )),
		self->gfx,
		self->gfx_size
	);

	memcpy(
		(cloned.vts = malloc( self->vts_size )),
	   	self->vts,
		self->vts_size
	);
	cloned.vts_ofs = (uintptr_t)cloned.vts - (uintptr_t)self->vts;

	return cloned;
}

typedef struct {
	StaticLayer*	layers;
	Material*		materials;
	Texture*		textures;
	uint16_t		num_layers;
	uint16_t		num_materials;
	uint16_t		num_textures;
	float			opacity;
} StaticMesh;

#define STATICMESH_INIT( lays, nlays, mtls, nmtls, texs, ntexs ) \
{ .layers = lays, .num_layers = nlays, .materials = mtls, .num_materials = nmtls, .textures = texs, .num_textures = ntexs, .opacity = 1.0f }

static inline StaticMesh clone_StaticMesh( StaticMesh const* self ) {
	StaticMesh cloned = *self;

	cloned.layers = malloc( sizeof(StaticLayer) * self->num_layers );
	for( size_t i=0; i<self->num_layers; ++i ) {
		cloned.layers[i] = clone_StaticLayer( &self->layers[i] );
	}

	cloned.materials = malloc( sizeof(Material) * self->num_materials );
	for( size_t i=0; i<self->num_materials; ++i ) {
		cloned.materials[i] = clone_Material( &self->materials[i] );
	}

	cloned.textures = malloc( sizeof(Texture) * self->num_textures );
	for( size_t i=0; i<self->num_textures; ++i ) {
		cloned.textures[i] = clone_Texture( &self->textures[i] );
	}

	return cloned;
}

static inline void destroy_StaticMesh( StaticMesh const* self ) {
	for( size_t i=0; i<self->num_textures; ++i ) {
		destroy_Texture( &self->textures[i] );
	}
	free( self->textures );

	for( size_t i=0; i<self->num_materials; ++i ) {
		destroy_Material( &self->materials[i] );
	}
	free( self->materials );

	for( size_t i=0; i<self->num_layers; ++i ) {
		destroy_StaticLayer( &self->layers[i] );
	}
	free( self->layers );
}

static inline void setOpacity_StaticMesh( StaticMesh* self, float const opacity ) {
	self->opacity = opacity;
}

static inline bool isOpaque_StaticMesh( StaticMesh const* self ) {
	return self->opacity >= 1.0f;
}

#endif	/* MYHEADER_H */
