#ifndef GFXDEVICE_H
#define GFXDEVICE_H

#include "dltrans.h"
#include "sys/microcode.h"
#include "nativedl.h"
#include "surface.h"
#include <types.h>

#define DEFAULT_BUFFER_W	320
#define DEFAULT_BUFFER_H	240

typedef enum {
	Draw_SurfaceType_GfxDevice,
	Display_SurfaceType_GfxDevice,
	Draw0_SurfaceType_GfxDevice,
	Draw1_SurfaceType_GfxDevice,
	Z_SurfaceType_GfxDevice,
	Aux_SurfaceType_GfxDevice
} SurfaceType_GfxDevice;

typedef enum {
	GFXCOLORMODE_RGBA_5551
} GfxColorMode;

typedef struct {
	uint16_t		width;
	uint16_t		height;
	GfxColorMode	colorMode;
} GfxMode;

typedef struct {
	uint32_t	modeIdx;
	OSMesgQueue	viMessageQ;
	OSMesg		viMessages;
	OSMesgQueue	vbiMessageQ;
	OSMesg		vbiMessages;
	OSMesgQueue	rspMessageQ;
	OSMesg		rspMessages;
	OSMesgQueue	rdpMessageQ;
	OSMesg		rdpMessages;
	bool		isRendering;
	DLTrans		dltrans;
	NativeDL	displayList;
	Microcode	microcode;
	OSTask		tlist;
	int			frameCount;

	/* surfaces */
	Surface*	colorBuffer;		/* surface to DRAW on */
	Surface		colorBuffer0;
	Surface		colorBuffer1;
	Surface		auxBuffer;
	Surface		zBuffer;

	/* RSP stuff. We use KSEG0 virtual adressing, assuming that RSP maps segment 0 to physical address 0x0 */
	u64			dram_stack[SP_DRAM_STACK_SIZE64] __attribute__((aligned (16)));
	u64			yieldBuffer[OS_YIELD_DATA_SIZE/sizeof(u64)];
} GfxDevice;

void			init_GfxDevice( GfxDevice* self );
Surface const*	getSurface_GfxDevice( GfxDevice const*, SurfaceType_GfxDevice const type );
void			swapBuffers_GfxDevice( GfxDevice* self );
uint32_t		getModeIndex_GfxDevice( GfxDevice const* self );
GfxMode const*	getCurrentMode_GfxDevice( GfxDevice const* self );
void			setMode_GfxDevice( GfxDevice* self, uint32_t const mode_idx );
GfxMode const*	enumerateModes_GfxDevice( GfxDevice const* self );
void			waitForVBI_GfxDevice( GfxDevice const* self );
void			waitForGPU_GfxDevice( GfxDevice const* self );

static inline void const* getMemoryEnd_GfxDevice( GfxDevice const* self ) {
	return getLastByte_Surface( &self->auxBuffer );
}

static inline bool isRendering_GfxDevice( GfxDevice const* self ) {
	return self->isRendering;
}

static inline NativeDL const* getNativeDLConst_GfxDevice( GfxDevice const* self ) {
	return &self->displayList;
}

static inline NativeDL* getNativeDL_GfxDevice( GfxDevice* self ) {
	return &self->displayList;
}

static inline Microcode const* getMicrocodeConst_GfxDevice( GfxDevice const* self ) {
	return &self->microcode;
}

static inline void setMicrocode_GfxDevice( GfxDevice* self, Microcode const microcode ) {
	self->microcode							= microcode;
	self->microcode.tlist.dram_stack		= &self->dram_stack[0];
	self->microcode.tlist.yield_data_ptr	= self->yieldBuffer;
}

static inline void setDefaultMicrocode_GfxDevice( GfxDevice* self ) {
	setMicrocode_GfxDevice( self,
		create_Microcode(
			F3DEX2_MicrocodeId,
			(u64*)gspF3DEX2_xbusTextStart,
			(u64*)gspF3DEX2_xbusDataStart
		)
	);
}

static inline float zAsFloat_GfxDevice( GfxDevice const* self, uint32_t const zval ) {
	if( self ) {}

	static struct {
		uint32_t shift;
		uint32_t add;
	} const z_tab[] = {
		{ 6, 0x00000 },
		{ 5, 0x20000 },
		{ 4, 0x30000 },
		{ 3, 0x38000 },
		{ 2, 0x3C000 },
		{ 1, 0x3E000 },
		{ 0, 0x3F000 },
		{ 0, 0x3F800 },
	};

	uint16_t const exponent = (zval >> 2) >> 11;
	uint16_t const mantissa = (zval >> 2) & 0x7FF;
	int32_t const decompressed = (mantissa << z_tab[exponent].shift) + z_tab[exponent].add;

	return ((float)((decompressed >> (1+2)) - 0x3FE0)) / 0x3FE0;
}

static inline size_t getTexelMemorySize_GfxDevice( GfxDevice const* self, bool const isTextureColorIndexed ) {
	if( self ) {}
	return 4096 / (isTextureColorIndexed ? 2 : 1);
}

void runGPU_GfxDevice( GfxDevice const* self, DList const* dlist );

#endif	/* GFXDEVICE_H */
