#ifndef TEXTURE_H
#define TEXTURE_H

#include <abort.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>
#include <types.h>

typedef struct {
	void const*	data;
	uint_t		referenceCount;
} Shared;

static inline Shared* new_Shared( void const* data ) {
	Shared* self = malloc( sizeof(Shared) );
	(*self) = (Shared) {
		.data = data,
		.referenceCount = 0
	};

	return self;
}

static inline Shared* borrow_Shared( Shared* self ) {
	self->referenceCount += 1;
	return self;
}

static inline void return_Shared( Shared* self ) {
	if( self->referenceCount > 0 ) {
		self->referenceCount -= 1;
		return;
	} else {
		abort2();
	}

	free( (void*)self->data );
}

static inline void delete_Shared( Shared* self ) {
	assert( 0 == self->referenceCount );
	free( self );
}

typedef enum {
	Ci4Ia16_TextureFormat,
	Ci4_TextureFormat,
	Rgba16_TextureFormat,
	Rgba32_TextureFormat,
	Unknown_TextureFormat,
	// below are deprecated names - fix lwo2ultra and remove them
	TEXTURE_FMT_CI4_IA16 = Ci4Ia16_TextureFormat,
	TEXTURE_FMT_CI4 = Ci4_TextureFormat,
	TEXTURE_FMT_RGBA16 = Rgba16_TextureFormat,
	TEXTURE_FMT_RGBA32 = Rgba32_TextureFormat,
	TEXTURE_FMT_UNKNOWN = Unknown_TextureFormat
} TextureFormat;

static inline bool isColorIndexed_TextureFormat( TextureFormat const self ) {
	switch( self ) {
		case Ci4Ia16_TextureFormat	: return true;
		case Ci4_TextureFormat		: return true;
		case Rgba16_TextureFormat	: return false;
		case Rgba32_TextureFormat	: return false;
		case Unknown_TextureFormat	: { abort(); }
		default: {
			abort();
		}
	}
}

static inline uint8_t getBytesPerPixel_TextureFormat( TextureFormat const self ) {
	switch( self ) {
		case Ci4Ia16_TextureFormat	: return 1;				// 1?
		case Ci4_TextureFormat		: return 1;				// should be 0.5 actually...
		case Rgba16_TextureFormat	: return 2;
		case Rgba32_TextureFormat	: return 4;
		case Unknown_TextureFormat	: { abort(); }
		default: {
			abort();
		}
	}
}

typedef struct {
	uint16_t 		width;
	uint16_t 		height;
	Shared*			pixels;
	Shared*			palette;
	void*			ownedPixels_;
	void const*		ownedPalette_;
	uint32_t		pixelsSize;
	uint16_t		paletteSize;
	TextureFormat	format;
} Texture;

/* static initialization */
#define TEXTURE_INIT(w, h, pix, pixsize, pal, palsize, fmt) { \
	.width = w, .height = h, \
	.pixels = NULL, \
	.ownedPixels_ = pix, \
	.ownedPalette_ = pal, \
	.pixelsSize = pixsize, \
	.paletteSize = palsize, \
	.format = fmt \
}

static inline Texture create_Texture(
	size_t const width, size_t const height,
	Shared* pixels,
	Shared* palette, size_t const paletteSize,
	TextureFormat const format
) {
	assert( pixels );

	return (Texture) {
		.width = width,
		.height = height,
		.pixels = borrow_Shared( pixels ),
		.palette = palette ? borrow_Shared( palette ) : palette,
		.ownedPixels_ = NULL,
		.ownedPalette_ = NULL,
		.pixelsSize = width * height * getBytesPerPixel_TextureFormat( format ),
		.paletteSize = paletteSize,
		.format = format
	};
}

static inline void destroy_Texture( Texture* self ) {
	assert( NULL == self->ownedPixels_ );
	assert( NULL == self->ownedPalette_ );

	return_Shared( self->pixels );
	self->palette ? return_Shared( self->palette ) : (void)0;
}

static inline uint32_t getPixelCount_Texture( Texture const* self ) {
	return self->width * self->height;
}

static inline uint16_t getWidth_Texture( Texture const* self ) {
	return self->width;
}

static inline uint16_t getHeight_Texture( Texture const* self ) {
	return self->height;
}

static inline TextureFormat getFormat_Texture( Texture const* self ) {
	return self->format;
}

static inline uint8_t getBytesPerPixel_Texture( Texture const* self ) {
	return getBytesPerPixel_TextureFormat( getFormat_Texture( self ) );
}

static inline size_t getSizeOfPixels_Texture( Texture const* self ) {
	return getBytesPerPixel_Texture( self ) * getPixelCount_Texture( self ) /
		(
			(
				Ci4Ia16_TextureFormat == getFormat_Texture( self ) ||
				Ci4_TextureFormat == getFormat_Texture( self )
			) ?
					2 : 1
		);
}
/*
static inline void* getPixels_Texture( Texture* self ) {
	return self->pixels;
}
*/
static inline uint16_t getPixelU16_Texture( Texture const* self, size_t const x, size_t const y ) {
	// TODO
	return ((uint16_t const*)self->pixels->data)[y * getWidth_Texture( self ) + x];
}

#if 0
static inline void setPixels_Texture( Texture* self, void* pixels ) {
	if( self ) {}
	if( pixels ) {}
	//self->pixels = pixels;
	// TODO!!!
}
#endif

/*
static inline void setPixelU16_Texture( Texture const* self, size_t const x, size_t const y, uint16_t const color ) {
	((uint16_t*)self->pixels)[y * getWidth_Texture( self ) + x] = color;
}
*/
static inline void const* getPixelsConst_Texture( Texture const* self ) {
	return self->pixels ? self->pixels->data : self->ownedPixels_;
}

static inline void const* getPaletteConst_Texture( Texture const* self ) {
	return self->palette ? self->palette->data : self->ownedPalette_;
}

static inline Texture copy_Texture( Texture const* source ) {
	Texture texture = create_Texture(
		getWidth_Texture( source ),
		getHeight_Texture( source ),
		source->pixels,
		source->palette,
		source->paletteSize,
		getFormat_Texture( source )
	);

	return texture;
}

static inline Texture clone_Texture( Texture const* source ) {
	void* pixels = malloc( source->pixelsSize );
	memcpy(
		pixels,
		source->pixels ? source->pixels->data : source->ownedPixels_,
		source->pixelsSize
	);

	void* palette = NULL;
	if( source->paletteSize ) {
		palette = malloc( source->paletteSize );
		memcpy(
			palette,
			source->palette ? source->palette->data : source->ownedPalette_,
			source->paletteSize
		);
	}

	Texture texture = create_Texture(
		getWidth_Texture( source ),
		getHeight_Texture( source ),
		new_Shared( pixels ),
		palette ? new_Shared( palette ) : NULL,
		source->paletteSize,
		getFormat_Texture( source )
	);

	return texture;
}

#endif	/* TEXTURE_H */

