#include <abort.h>
#include <assert.h>
#include <float.h>
#include <engine.h>
#include <rectangle.h>
#include <shadow.h>
#include <sys/region.h>
#include <sys/sys.h>
#include <node/xformer_funcs.h>
#include <containers/iterator_ext.h>

#define MAX_XFORMERS		256		//+256
#define MAX_STATIC_MESHES	32
#define MAX_MESHES			256		//+256
#define MAX_BILLBOARDS		32
#define MAX_CAMERAS			32
#define MAX_LUXES			32
#define MAX_MATERIALS		256

static FragmentShader const
flareFragmentShader = {
	.numberOfPasses = 1,
	.pass[0] = {
		G_CCMUX_TEXEL0, G_CCMUX_0, G_CCMUX_PRIMITIVE, G_CCMUX_0,
		G_ACMUX_TEXEL0, G_ACMUX_0, G_ACMUX_PRIMITIVE, G_ACMUX_0
	},
},
shadeOpaqueFragmentShader = {
	.numberOfPasses = 1,
	.pass[0] = {
		G_CCMUX_SHADE, G_CCMUX_0, G_CCMUX_PRIMITIVE, G_CCMUX_0,
		G_ACMUX_0, G_ACMUX_0, G_ACMUX_0, G_ACMUX_PRIMITIVE
	},
},
decalOpaqueFragmentShader = {
	.numberOfPasses = 1,			// 2
	.pass[0] = {
		G_CCMUX_0, G_CCMUX_0, G_CCMUX_0, G_CCMUX_TEXEL0,	//G_CCMUX_1, G_CCMUX_TEXEL0, G_CCMUX_ENVIRONMENT, G_CCMUX_0,
		G_ACMUX_0, G_ACMUX_0, G_ACMUX_0, G_ACMUX_TEXEL0
	},
	//
	.pass[1] = {
		G_CCMUX_TEXEL0, G_CCMUX_0, G_CCMUX_COMBINED, G_CCMUX_0,
		G_ACMUX_0, G_ACMUX_0, G_ACMUX_0, G_ACMUX_TEXEL0
	},
},
decalTranslucentFragmentShader = {
	.numberOfPasses = 1,		// 2
	.pass[0] = {
		G_CCMUX_0, G_CCMUX_0, G_CCMUX_0, G_CCMUX_TEXEL0,
		G_ACMUX_0, G_ACMUX_0, G_ACMUX_0, G_ACMUX_PRIMITIVE
	},
	//
	.pass[1] = {
		G_CCMUX_TEXEL0, G_CCMUX_0, G_CCMUX_COMBINED, G_CCMUX_0,
		G_ACMUX_0, G_ACMUX_0, G_ACMUX_0, G_ACMUX_PRIMITIVE
	},
};

void init_Engine( Engine* self ) {
	init_DList( &self->dl );

	init_Vec( &self->lights,
		sizeof(Lux const*),
		Default_RegionAlignment,
		NULL
	);
	init_Vec( &self->materials,
		sizeof(Material const*),
		Default_RegionAlignment,
		NULL
	);
	init_Vec( &self->textures,
		sizeof(Texture const*),
		Default_RegionAlignment,
		NULL
	);
	init_Vec( &self->transparentXf,
		sizeof(Raw_Xformer*),
		Default_RegionAlignment,
		NULL
	);
	init_Vec( &self->opaqueXf,
		sizeof(Raw_Xformer*),
		Default_RegionAlignment,
		NULL
	);
	init_Vec( &self->skyboxesXf,
		sizeof(Skybox_Xformer*),
		Default_RegionAlignment,
		NULL
	);
	init_Vec( &self->cpuMatrixes,
		sizeof(Matrix),
		B4_RegionAlignment,
		NULL
	);

	init_MemPool( &self->xformersPool,
		sizeof(Xformer),
		MAX_XFORMERS,
		Default_RegionAlignment
	);		// how to get an alignment of a structure?
	init_MemPool( &self->staticMeshesPool,
		sizeof(StaticMesh),
		MAX_STATIC_MESHES,
		Default_RegionAlignment
	);
	init_MemPool( &self->meshesPool,
		sizeof(Mesh),
		MAX_MESHES,
		Default_RegionAlignment
	);
	init_MemPool( &self->billboardsPool,
		sizeof(Billboard),
		MAX_BILLBOARDS,
		Default_RegionAlignment
	);
	init_MemPool( &self->camerasPool,
		sizeof(Camera),
		MAX_CAMERAS,
		Default_RegionAlignment
	);
	init_MemPool( &self->luxesPool,
		sizeof(Lux),
		MAX_LUXES,
		Default_RegionAlignment
	);
	init_MemPool( &self->materialsPool,
		sizeof(Material),
		MAX_MATERIALS,
		Default_RegionAlignment
	);

	setViewport_Engine( self, create_Viewport( 0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 1.0f ) );			// TODO: not sure about the last two values
	setZBuffer_Engine( self, NULL );

	self->camera2dXf = createCamera_Engine( self, Orthographic_CameraType );

	static Material default_material = MATERIAL_INIT(
		VECTOR4F_SET( 1.0f, 1.0f, 1.0f, 1.0f ),
		Const_TextureMapping,
		Decal_TextureCombine,
		TEXTURE_ID_NONE
	);
	setLighting_Material( &default_material, true );
	addMaterialReference_Engine( self, &default_material );
}

static inline Matrix* getProjectionMatrix_Engine( Engine* self ) {
	return &self->projMtxF;
}

static void handleGeometryMode( DList* dlistOut, Material const* material ) {
	if( isLighted_Material( material ) ) {
		pushSetGeometryMode_DList( dlistOut, Lighting_GeometryMode_DList );
	} else {
		pushClearGeometryMode_DList( dlistOut, Lighting_GeometryMode_DList );
	}

	switch( getShadingType_Material( material ) ) {
		case Flat_ShadingType_Material:
			pushClearGeometryMode_DList( dlistOut, Gouraud_GeometryMode_DList );
			break;
		case Smooth_ShadingType_Material:
			pushSetGeometryMode_DList( dlistOut, Gouraud_GeometryMode_DList );
			break;
		default:
			abort();
	}
}

static void handleRenderingModes(
	DList* dlistOut,
	Material const* material,
	Texture const* textures[],
	Rectangle const* tile,
	Type_Primitives_Layer const primitiveType
) {
	ActiveTexturesMask_DList	activeTextureMask;
	RenderMode_DList			renderMode;
	FragmentShader const*		fragmentShader = &shadeOpaqueFragmentShader;

	if( TEXTURE_ID_NONE == getTextureId_Material( material ) ) {
		activeTextureMask = 0;

		renderMode = (getColorConst_Material( material )->t == 1.0f) ?
			Opaque_RenderMode_DList :		//Z_TexturedEdge_RenderMode_DList :
			Translucent_RenderMode_DList;	// ? ZB_OPAQUE

		fragmentShader = &shadeOpaqueFragmentShader;
	} else {
		activeTextureMask = bit0_ActiveTexturesMask_DList;

		Texture const* tex = textures[ getTextureId_Material( material ) ];
		Rectangle const tttile = {
			0,
			0,
			getWidth_Texture( tex ) - 1,
			getHeight_Texture( tex ) - 1
		};

		//Rectangle const* tile = &layer->tile;
		if( (tile->p0x == 0) && (tile->p0y == 0) && (tile->p1x == 0) && (tile->p0y == 0) ) {		// HACKY
			pushLoadTexture_DList( dlistOut, tex, tttile, 0, false );
		} else {
			pushLoadTexture_DList( dlistOut, tex, *tile, 0, true );
		}
		#if 1
		if( TEXTURE_ID_NONE != material->texture1Id ) {
			activeTextureMask |= bit1_ActiveTexturesMask_DList;

			Texture const* tex = textures[ material->texture1Id ];
			Rectangle const tttile = {
				0,
				0,
				getWidth_Texture( tex ) - 1,
				getHeight_Texture( tex ) - 1
			};

			//Rectangle const* tile = &layer->tile;
			if( (tile->p0x == 0) && (tile->p0y == 0) && (tile->p1x == 0) && (tile->p0y == 0) ) {
				pushLoadTexture_DList( dlistOut, tex, tttile, 1, false );
			} else {
				pushLoadTexture_DList( dlistOut, tex, *tile, 1, true );		// !
			}
		}
		#endif

		switch( getTextureCombine_Material( material ) ) {
			case Decal_TextureCombine: {
				renderMode = (getColorConst_Material( material )->t == 1.0f) ?
					TexturedEdge_RenderMode_DList :
					Translucent_RenderMode_DList;	// ? ZB_OPAQUE

				fragmentShader = (getColorConst_Material( material )->t == 1.0f) ?
					&decalOpaqueFragmentShader :
					&decalTranslucentFragmentShader;
				break;
			}
			case Alpha_TextureCombine: {
				renderMode = Translucent_RenderMode_DList;
				fragmentShader = &flareFragmentShader;
				break;
			}
			default: {
				abort();
			}
		}

		/*
		if( Env_TextureMapping == getTextureMapping_Material( &material ) ) {
			pushSetTextureScale_DList( dlist_out, 0.125f, 0.125f );
			pushSetGeometryMode_DList( dlist_out, Lighting_GeometryMode_DList | TextureGen_GeometryMode_DList );
			Coordsys const* camcs = getCoordsys_Xformer( cameraXf->xf );
			Coordsys const* meshcs = getCoordsys_Xformer( staticMeshXf->xf );
			
			pushSetLookAt_DList( dlist_out, &camcs->origin, &meshcs->origin, &camcs->v );
		}
		*/
		
		//temp: seems there is no render mode for non-transparent lines?
		if( Segment_PrimitiveType_Layer == primitiveType ) {
			renderMode = Translucent_RenderMode_DList;
			fragmentShader = &decalTranslucentFragmentShader;
		}
	}

	pushSetFragmentShader_DList( dlistOut, fragmentShader );
	pushSetActiveTexturesMask_DList( dlistOut, activeTextureMask );
	pushSetRenderMode_DList( dlistOut, renderMode );
}

static void handleStaticMesh(
	DList*						dlist_out,
	StaticMesh_Xformer const*	staticMeshXf,
	Camera_Xformer const*		cameraXf
) {
	if( cameraXf ) {}

	////////////////////// TODO
	pushSetGeometryMode_DList( dlist_out,
		Lighting_GeometryMode_DList |
		Shade_GeometryMode_DList 
	);
	pushLoadMatrix_DList( dlist_out,
		getResultMatrix_Xformer( staticMeshXf->xf ),
		Modelview_MatrixType_DList
	);

	StaticMesh const* staticMesh = staticMeshXf->staticMesh;
	Texture const* textures[staticMesh->num_textures + 1];		// TODO: check how standard treats a 0-sized array
	for( size_t textureIndex = 0; textureIndex < staticMesh->num_textures; ++textureIndex ) {
		textures[textureIndex] = &staticMesh->textures[textureIndex];
	}
			
	StaticLayer const* layer = staticMesh->layers;
	for( size_t i=0; i<staticMesh->num_layers; ++i, ++layer ) {
		Material material = clone_Material( &staticMesh->materials[ layer->material_id ] );
		getColor_Material( &material )->t *= staticMesh->opacity;
		pushSetMaterial_DList( dlist_out, &material );

		handleGeometryMode( dlist_out, &material );
		handleRenderingModes( dlist_out, &material, textures, &layer->tile, Triangle_PrimitiveType_Layer );

		pushSetSegment_DList( dlist_out, 0x0, (void*)layer->vts_ofs );
		pushJumpToTranslated_DList( dlist_out, layer->gfx );
	}
}

// TODO: Group objects that share a single mesh and load vertex buffer only once.
static void handleMesh(
	DList*					dlist_out,
	Mesh_Xformer const*		meshXf,
	Camera_Xformer const*	cameraXf,
	Material const*			materials[],
	Texture const*			textures[]
) {
	if( cameraXf ) {}
	pushSetVertexShader_DList( dlist_out,
		meshXf->vertexShader
	);
	////////////////////// TODO
	pushSetGeometryMode_DList( dlist_out,
		Lighting_GeometryMode_DList |
		Shade_GeometryMode_DList
	);
	pushLoadMatrix_DList( dlist_out,
		getResultMatrix_Xformer( meshXf->xf ),
		Modelview_MatrixType_DList
	);


	VecConstIterator layersIter =
		createConstIterator_Vec( getLayersConst_Mesh( meshXf->mesh ),
			(IteratorRange) { .type = All_IteratorRange }
		);
	size_t layerIdx = 0;
	for( Layer const* layer;
		(layer = next_VecConstIterator( &layersIter ));
		++layerIdx
	) {
		Material material = clone_Material( materials[ getMaterialId_Layer( layer ) ] );
		getColor_Material( &material )->t *= getOpacity_Mesh( meshXf->mesh );
		pushSetMaterial_DList( dlist_out, &material );

		handleGeometryMode( dlist_out, &material );
		handleRenderingModes( dlist_out, &material, textures, &layer->tile, getPrimitiveType_Layer( layer ) );

		{
			Vec const* verts = getVerticesConst_Layer( layer );
			pushSetVertexSource_DList( dlist_out,
				getHeadConst_Vec( verts ),
				0
			);
			pushLoadVertexBuffer_DList( dlist_out,
				0,
				0,
				getLength_Vec( verts )
			);
		}

		{
			PrimitiveType_DList primitiveType;
			switch( getPrimitiveType_Layer( layer ) ) {
				case Triangle_PrimitiveType_Layer:
					primitiveType = Triangle_PrimitiveType_DList;
					break;
				case Segment_PrimitiveType_Layer:
					primitiveType = Segment_PrimitiveType_DList;
					break;
				default:
					abort();
			}
			Vec const* primitives = &getPrimitivesConst_Layer( layer )->primitives;
			pushSetPrimitiveSource_DList( dlist_out,
				primitiveType,
				getHeadConst_Vec( primitives )
			);
			pushDrawPrimitives_DList( dlist_out,
				primitiveType,
				getLength_Vec( primitives ),
				0	//layer_i*3
			);
		}
	}
	destroy_VecConstIterator( &layersIter );
}

static void handleBillboard( DList* dlist, Billboard_Xformer const* billboardXf, Camera_Xformer const* cameraXf ) {
	handleStaticMesh( dlist, (StaticMesh_Xformer*)billboardXf, cameraXf );
}

static void handleSkybox(
	DList*					dlist_out,
	Skybox_Xformer const*	skyboxXf,
	Camera_Xformer const*	cameraXf,
	Material const*			materials[],
	Texture const*			textures[]
) {
	Mesh_Xformer meshXf = {
		.tag			= Mesh_ObjectType,
		.xf				= skyboxXf->xf,
		.mesh			= skyboxXf->skybox->mesh,
		.vertexShader	= NULL
	};

	handleMesh( dlist_out, &meshXf, cameraXf, materials, textures );
}

static void handleProjection( DList* dlist, Matrix* projectionMtx, Viewport const* viewport, float const screenAspectRatio, Camera_Xformer const* cameraXf ) {
	switch( getType_Camera( cameraXf->camera ) ) {
		case Perspective_CameraType: {
			PerspCamera const* camera = getPersp_Camera( cameraXf->camera );
			float perspNorm;
			perspective_Matrix( projectionMtx,
				&perspNorm,
				getFovY_PerspCamera( camera ),
				screenAspectRatio * getAspectRatio_Viewport( viewport ),
				getNear_PerspCamera( camera ),
				getFar_PerspCamera( camera )
			);
			pushNormalizePerspective_DList( dlist, perspNorm );

			break;
		}
		case Orthographic_CameraType: {
			OrthoCamera const* camera = getOrtho_Camera( cameraXf->camera );
			orthographic_Matrix( projectionMtx,
				getLeft_OrthoCamera( camera ),
				getRight_OrthoCamera( camera ),
				getBottom_OrthoCamera( camera ),
				getTop_OrthoCamera( camera ),
				getNear_OrthoCamera( camera ),
				getFar_OrthoCamera( camera )
			);
			break;
		}
		default:
			abort();
			break;
	}

	Matrix inv;
	invert_Matrix( getMatrix_Xformer( cameraXf->xf ), &inv );
	multiply_Matrix( projectionMtx, &inv, projectionMtx );
	pushLoadMatrix_DList( dlist, projectionMtx, Projection_MatrixType_DList );
}

static bool light_filter( void const* element_generic ) {
	TreeNode const* tree_node = *(TreeNode const**)element_generic;
	Raw_Xformer const* rawXf = getData_TreeNode( tree_node );

	return Lux_ObjectType == rawXf->tag;
}

static void light_map( void* dest_, void const* src_ ) {
	Raw_Xformer const** dest = (Raw_Xformer const**)dest_; 
	TreeNode*** src = (TreeNode***)src_;

	*dest = getData_TreeNode( **src );
}

static Vec collectLights( Vec const* tn_vec ) {
	VecConstIterator it = createConstIterator_Vec( tn_vec, (IteratorRange) { .type = All_IteratorRange } );
	Vec filtered = filter_IteratorConst( &it.base, light_filter );

	VecConstIterator lights_it = createConstIterator_Vec( &filtered ,  (IteratorRange) { .type = All_IteratorRange } );
	Vec mapped = map_IteratorConst(
		&lights_it.base,
		&(Object) { .size_of = sizeof(Raw_Xformer*), .alignment = Default_RegionAlignment, .clone_func = NULL, .destroy_func = NULL },
		light_map
	);
	destroy_Vec( &filtered );

	return mapped;
}

static bool flare_filter( void const* element_generic ) {
	Raw_Xformer const* rawXf = *(Raw_Xformer const**)element_generic;

	return Flare_ObjectType == rawXf->tag;
}

static void flare_map( void* dest_, void const* src_ ) {
	Raw_Xformer** dest = (Raw_Xformer**)dest_; 
	Raw_Xformer*** src = (Raw_Xformer***)src_;

	*dest = **src;
}

static Vec collectFlares( Vec const* tn_vec ) {
	VecConstIterator it = createConstIterator_Vec( tn_vec, (IteratorRange) { .type = All_IteratorRange } );
	Vec filtered = filter_IteratorConst( &it.base, flare_filter );

	VecConstIterator flares_it = createConstIterator_Vec( &filtered ,  (IteratorRange) { .type = All_IteratorRange } );
	Vec mapped = map_IteratorConst(
		&flares_it.base,
		&(Object) { .size_of = sizeof(Raw_Xformer*), .alignment = Default_RegionAlignment, .clone_func = NULL, .destroy_func = NULL },
		flare_map
	);
	destroy_Vec( &filtered );

	return mapped;
}

// TODO
#if 0
static void splitByPrimitiveType( Vec* withTrianglesXf, Vec* withSegmentsXf, Vec const* tn_vec ) {
	VecConstIterator it = createConstIterator_Vec( tn_vec, (IteratorRange) { .type = ITERATORRANGE_ALL } );
	for( Raw_Xformer const* const* raw_xf; (raw_xf = next_VecConstIterator( &it )); ) {
		switch( (*raw_xf)->tag ) {
			case OBJECT_MESH: {
				Mesh const* mesh = upcast_Mesh_Xformer( *raw_xf )->mesh;
				switch( getPrimitiveType_Layer( getLayerConst_Mesh( mesh, 0 ) ) ) {
					case		Triangle_PrimitiveType_Layer:	break;
					case		Segment_PrimitiveType_Layer:	break;
					default:	abort();
				}
				break;
			}
			default:
				break;
		}
	}
	destroy_VecConstIterator( &it );
}
#endif

// NO NULL
#if 0
static void collectShadowRelated( Vec* shadowCastersXf, Vec* shadowReceiversXf, TreeNode const* tn_list[] ) {
	for( TreeNode const** it = tn_list; NULL != *it; ++it ) {
		Raw_Xformer const* raw_xf = getData_TreeNode( *it );
		switch( raw_xf->tag ) {
			case OBJECT_MESH: {
				Mesh_Xformer const* mesh_xf = (Mesh_Xformer const*)raw_xf;
				MeshAttribute attributes = getAttributes_Mesh( mesh_xf->mesh );
				if( MESHATTRIBUTE_CASTS_SHADOW & attributes ) {
					pushBack_Vec( shadowCastersXf, mesh_xf );
				}
				if( MESHATTRIBUTE_RECEIVES_SHADOW & attributes ) {
					pushBack_Vec( shadowReceiversXf, mesh_xf );
				}
				break;
			}
			default:
				break;
		}
	}
}
#endif

// NO NULL
static void splitOpaqueTransparentSkyboxes( Vec* opaqueXf, Vec* transparentXf, Vec* skyboxesXf, Vec const* tn_vec ) {
	VecConstIterator it = createConstIterator_Vec( tn_vec, (IteratorRange) { .type = All_IteratorRange } );
	for( Raw_Xformer const* const* rawXfRef; (rawXfRef = next_VecConstIterator( &it )); ) {
		Raw_Xformer const* rawXf = *rawXfRef;
		switch( rawXf->tag ) {
			case Billboard_ObjectType: {
				Billboard const* billboard = ((Billboard_Xformer*)rawXf)->billboard;
				isOpaque_StaticMesh( &billboard->mesh ) ?
					pushBack_Vec( opaqueXf, &rawXf ) :
					pushBack_Vec( transparentXf, &rawXf );
				break;
			}
			case StaticMesh_ObjectType: {
				StaticMesh const* staticMesh = ((StaticMesh_Xformer*)rawXf)->staticMesh;
				isOpaque_StaticMesh( staticMesh ) ?
					pushBack_Vec( opaqueXf, &rawXf ) :
					pushBack_Vec( transparentXf, &rawXf );
				break;
			}
			case Skybox_ObjectType: {
				pushBack_Vec( skyboxesXf, &rawXf );
				break;
			}
			case Mesh_ObjectType: {
				Mesh const* mesh = ((Mesh_Xformer*)rawXf)->mesh;
				isOpaque_Mesh( mesh ) ?
					pushBack_Vec( opaqueXf, &rawXf ) :
					pushBack_Vec( transparentXf, &rawXf );
				break;
			}
			default:
				break;
		}
	}
	destroy_VecConstIterator( &it );
}

static void calculateWorldCoordinates( Matrix cpuMtx[], Vec const* tn_vec ) {
	Matrix* matrix_it = cpuMtx;
	VecConstIterator it = createConstIterator_Vec( tn_vec, (IteratorRange) { .type = All_IteratorRange } );
	for( TreeNode const* const* tree_node; (tree_node = next_VecConstIterator( &it )); ) {
		TreeNode const* parent = getParent_TreeNode( *tree_node );

		finish_Xformer(
			( (Raw_Xformer*)getData_TreeNode( *tree_node ) )->xf,
			parent ?
				( (Raw_Xformer*)getData_TreeNode( parent ) )->xf
				: NULL,
			matrix_it
		);

		++matrix_it;
	}
	destroy_VecConstIterator( &it );
}

static void handleLighting( DList* dlist, Vec const* lightsXf ) {
	VecConstIterator lights = createConstIterator_Vec( lightsXf, (IteratorRange) { .type = All_IteratorRange } );
	for( Lux_Xformer const* const* luxXf; (luxXf = next_VecConstIterator( &lights )); ) {
		pushLoadLight_DList( dlist,
			( *luxXf )->lux,
			&getCoordsys_Xformer( (*luxXf)->xf )->origin,
			//&getCoordsys_Xformer( rawXf->xf )->n,
			0
		);
	}
	destroy_VecConstIterator( &lights );
}

#if 0
static Mesh castShadow( Mesh_Xformer const* caster, Mesh_Xformer const* receiver, Lux_Xformer const* light ) {
	if( caster ) {}
	if( receiver ) {}
	if( light ) {}

	Mesh shadow;
	init_Mesh( &shadow );

	return shadow;
}
#endif

// provide a destroy_func!!!!
#if 0
static void shadowMeshesXf_DestroyFunc( Mesh_Xformer* shadow_xf, Engine* engine ) {
	freeMesh_Engine( engine, shadow_xf );
}

static Vec handleShadows_Engine( Engine* self, Vec const* shadowCastersXf, Vec const* shadowReceiversXf, Vec const* lightsXf ) {
	Vec shadowMeshesXf;
	init_Vec( &shadowMeshesXf,
		sizeof(Mesh_Xformer),
		Default_RegionAlignment,
		NULL						// shadowMeshesXf_DestroyFunc
	);

	Mesh_Xformer const* const* receiver;
	VecConstIterator receivers = createConstIterator_Vec( shadowReceiversXf, (IteratorRange) { .type = ITERATORRANGE_ALL } );
	// iterate receivers
	while( (receiver = next_VecConstIterator( &receivers )) ) {
		Mesh_Xformer const* const* caster;
		VecConstIterator casters = createConstIterator_Vec( shadowCastersXf, (IteratorRange) { .type = ITERATORRANGE_ALL } );
		// iterate casters
		while( (caster = next_VecConstIterator( &casters )) ) {
			Lux_Xformer const* const* light;
			VecConstIterator lights = createConstIterator_Vec( lightsXf, (IteratorRange) { .type = ITERATORRANGE_ALL } );
			// iterate light sources
			while( (light = next_VecConstIterator( &lights ) ) ) {
				Mesh shadow = castShadow( *caster, *receiver, *light );
				Mesh_Xformer shadow_xf = createMesh_Engine( self );
				destroy_Mesh( shadow_xf.mesh );
				*shadow_xf.mesh = shadow;

				pushBack_Vec( &shadowMeshesXf, &shadow_xf );
			}
			destroy_VecConstIterator( &lights );
		}
		destroy_VecConstIterator( &casters );
	}

	destroy_VecConstIterator( &receivers );

	return shadowMeshesXf;
}
#endif

// NO NULL
static void applyAdditionalTransformations( Camera_Xformer const* cameraXf, Vec const* tn_vec ) {
	VecConstIterator it = createConstIterator_Vec( tn_vec, (IteratorRange) { .type = All_IteratorRange } );
	for( Raw_Xformer const* const* rawXf; (rawXf = next_VecConstIterator( &it )); ) {
		void* data;
		switch( (*rawXf)->tag ) {
			case Skybox_ObjectType: {
				data = cameraXf->xf;
				break;
			}
			case Billboard_ObjectType: {
				data = cameraXf->xf;
				break;
			}
			default: {
				data = NULL;
				break;
			}
		}

		process_Xformer( (*rawXf)->xf, data );
		reset_Xformer( (*rawXf)->xf );
	}
	destroy_VecConstIterator( &it );
}

static void buildSceneDl(
	DList*					dlist_out,
	Camera_Xformer const*	cameraXf,
	size_t const			groupsNum,
	Vec const*				xformersGrouped[groupsNum],
	Material const*			materials[],
	Texture const*			textures[]
) {
	for( size_t groupIdx=0; groupIdx<groupsNum; ++groupIdx ) {
		pushSetTextureScale_DList( dlist_out, 1.0f, 1.0f );

		VecConstIterator rawXfIter =
			createConstIterator_Vec( xformersGrouped[groupIdx],
				(IteratorRange) { .type = All_IteratorRange }
			);
		for(
			Raw_Xformer const* const* rawXf;
			(rawXf = (Raw_Xformer const* const*)next_VecConstIterator( &rawXfIter ));
		) {
			switch( (*rawXf)->tag ) {
				case Skybox_ObjectType: {
					//pushPushState_DList( dlist_out );
					handleSkybox(
						dlist_out,
						(Skybox_Xformer const*)*rawXf,
						cameraXf,
						materials,
				   		textures
					);
					//pushPopState_DList( dlist_out );
					break;
				}
				case Billboard_ObjectType: {
					handleBillboard( dlist_out, (Billboard_Xformer const*)*rawXf, cameraXf );
					break;
				}
				case StaticMesh_ObjectType: {
					handleStaticMesh( dlist_out, (StaticMesh_Xformer const*)*rawXf, cameraXf );
					break;
				}
				case Mesh_ObjectType: {
					handleMesh(
						dlist_out,
						(Mesh_Xformer const*)*rawXf,
						cameraXf,
						materials,
						textures
					);
					break;
				}
				default:
					abort();
					break;
			}
		}
		destroy_VecConstIterator( &rawXfIter );
	}
}

static Vec flattenTree_NO_NULL( TreeNode const* tree ) {
	Vec flatten = create_Vec( sizeof(Raw_Xformer*), Default_RegionAlignment, NULL );

	TreeNodeIterator ni = create_TreeNodeIterator( tree );
	for( TreeNode const* tn; (tn = inc_TreeNodeIterator( &ni )); ) {
		pushBack_Vec( &flatten, &(Raw_Xformer const*) { getData_TreeNode( tn ) } );
	};
	destroy_TreeNodeIterator( &ni );

	return flatten;
}

static Vec flattenTree( TreeNode const* tree ) {
	Vec flatten = create_Vec( sizeof(TreeNode*), Default_RegionAlignment, NULL );

	TreeNodeIterator ni = create_TreeNodeIterator( tree );
	for( TreeNode const* tn; (tn = inc_TreeNodeIterator( &ni )); ) {
		pushBack_Vec( &flatten, &tn );
	};
	destroy_TreeNodeIterator( &ni );

	return flatten;
}

void sceneTaskF_Engine( Engine* self, Tree const* scene ) {

	Vec tn_list = flattenTree( getRootConst_Tree( scene ) );					// Vec(TreeNode const*)
	Vec tn_list_nn = flattenTree_NO_NULL( getRootConst_Tree( scene ) );			// Vec(TreeNode const*)

	GfxMode const* gfxMode = getCurrentMode_GfxDevice( getGfxDevice_Sys() );
	handleProjection(
		getDisplayList_Engine( self ),
		getProjectionMatrix_Engine( self ),
		getViewport_Engine( self ),
		(float)gfxMode->width / (float)gfxMode->height,
		getCamera_Engine( self )
	);

	resize_Vec( &self->cpuMatrixes, getLength_Vec( &tn_list ) );
	calculateWorldCoordinates(
		getHead_Vec( &self->cpuMatrixes ),
		&tn_list
	);

	Vec lightsXf = collectLights( &tn_list );
	handleLighting( &self->dl, &lightsXf );

	applyAdditionalTransformations(
		getCamera_Engine( self ),
		&tn_list_nn
	);

	/*
	reset_Vec( &self->shadowCastersXf );
	reset_Vec( &self->shadowReceiversXf );
	collectShadowRelated(
		&self->shadowCastersXf,
		&self->shadowReceiversXf,
		(TreeNode const**)getHeadConst_Vec( &tn_list )
	);
	*/

	//Vec shadowMeshesXf = handleShadows_Engine( self, &self->shadowCastersXf, &self->shadowReceiversXf, &lightsXf );

	reset_Vec( &self->opaqueXf );
	reset_Vec( &self->transparentXf );
	reset_Vec( &self->skyboxesXf );
	splitOpaqueTransparentSkyboxes(
		&self->opaqueXf,
		&self->transparentXf,
		&self->skyboxesXf,
		&tn_list_nn
	);

	Vec const* xformersGrouped[] = { &self->skyboxesXf, &self->opaqueXf, &self->transparentXf };
	buildSceneDl(
		&self->dl,
		getCamera_Engine( self ),
		sizeof(xformersGrouped) / sizeof(*xformersGrouped),
		xformersGrouped,
		(Material const**)getHeadConst_Vec( &self->materials ),
		(Texture const**)getHeadConst_Vec( &self->textures )
	);

	//destroy_Vec( &shadowMeshesXf );
	destroy_Vec( &lightsXf );
	destroy_Vec( &tn_list_nn );
	destroy_Vec( &tn_list );
}

static void calculateFlarePositions( Flare_Xformer** flareXf, Vec* flare_nodes, TreeNode* root_node, Vector3f const* light_pos_ndc ) {
	GfxDevice const* gdev = getGfxDevice_Sys();
	float const aspect_ratio = getAspectRatio_Surface(
		getSurface_GfxDevice( gdev, Draw_SurfaceType_GfxDevice )
	);

	float const scale_by_view_angle = normalize_Vector3f( &(Vector3f) {
		.x = light_pos_ndc->x,
		.y = light_pos_ndc->y,
		.z = 0.5f
	}).z;
	Vector3f const final_pos_ndc = create_Vector3f( -light_pos_ndc->x, -light_pos_ndc->y, light_pos_ndc->z );
	Vector3f const dir = sub_Vector3f( &final_pos_ndc, light_pos_ndc );

	VecIterator pairs_iterator = createIterator_Vec( (*flareXf)->meshes, (IteratorRange) { .type = All_IteratorRange } );
	size_t pair_i = 0;
	for(
		Pair_StaticMesh_Xformer_Scale* pair_xf_s;
		(pair_xf_s = next_VecIterator( &pairs_iterator ));
		++pair_i
	) {
		{
			TreeNode tn = create_TreeNode( &(pair_xf_s->staticMeshXf), NULL, NULL, 0 );
			pushBack_Vec( flare_nodes, &tn );
		}
		attachChild_TreeNode( root_node, *(TreeNode*)getLast_Vec( flare_nodes ) );
		//attachChild_TreeNode( root_node, create_TreeNode( &(pair_xf_s->static_mesh_xf), NULL, NULL, 0 ) );

		// calculate intermediate flare
		size_t const pairs_num = getLength_Vec( (*flareXf)->meshes );
		Vector3f const curr_pos_ndc = (1 == pairs_num) ?
			*light_pos_ndc :
			add_Vector3f( light_pos_ndc,
				&VECTOR3F_TEMP( scale_Vector3f( &dir, pair_i * (1.0f / (pairs_num - 1)) ) )
			);

		// transform
		Matrix mv;
		scale_Matrix( &mv,
			&(Vector3f) {
				pair_xf_s->scale * scale_by_view_angle,
				aspect_ratio*pair_xf_s->scale * scale_by_view_angle,
				1.0f
			}
		);
		transformLocally_Xformer( pair_xf_s->staticMeshXf.xf, &mv );

		translate_Matrix( &mv, curr_pos_ndc.x, curr_pos_ndc.y, curr_pos_ndc.z );
		transformLocally_Xformer( pair_xf_s->staticMeshXf.xf, &mv );
	}
	destroy_VecIterator( &pairs_iterator );
}

static void renderFlares( Engine* engine, TreeNode root_node ) {
	Camera_Xformer const* cameraXf = getCamera_Engine( engine );
	setCamera_Engine( engine, &engine->camera2dXf );
	setParams_OrthoCamera(
		getOrtho_Camera( getCamera_Engine( engine )->camera ),
		-1.0f, 1.0f, -1.0f, 1.0f, 1.33f, 1000.0f
	);

	Tree flare_scene = create_Tree( root_node, NULL, NULL );
	sceneTaskF_Engine( engine, &flare_scene );
	destroy_Tree( &flare_scene );

	setCamera_Engine( engine, cameraXf );
}

// DANGER: must be invoked AFTER sceneTaskF already rendered a scene
void flareTaskF_Engine( Engine* self, Tree const* scene ) {
	if( NULL == getZBuffer_Engine( self ) ) { return; }

	Raw_Xformer root = createRaw_Engine( self );
	TreeNode root_node = create_TreeNode( &root, NULL, NULL, 0 );

	Vec tn_list = flattenTree_NO_NULL( getRootConst_Tree( scene ) );		// Vec(TreeNode*)
	Vec flaresXf = collectFlares( &tn_list );								// Vec(Raw_Xformer*)
	
	Vec flare_nodes = create_Vec( sizeof(TreeNode), Default_RegionAlignment, destroy_TreeNode_ );

	VecIterator flares_iterator = createIterator_Vec( &flaresXf, (IteratorRange) { .type = All_IteratorRange } );
	for( Flare_Xformer** flareXf; (flareXf = next_VecIterator( &flares_iterator )); ) {
		// calculate light position in clip coordinates
		Matrix mtx;
		multiply_Matrix( &mtx,
			getResultMatrix_Xformer( (*flareXf)->xf ),
			getProjectionMatrix_Engine( self )								// DANGER: assuming it is multiplied by the inverse view!!!
		);

		// clip coords -> NDC
		Vector3f const light_pos_ndc = create_Vector3f(
			MATRIX_GET(mtx, 3, 0) / MATRIX_GET(mtx, 3, 3),
			MATRIX_GET(mtx, 3, 1) / MATRIX_GET(mtx, 3, 3),
			MATRIX_GET(mtx, 3, 2) / MATRIX_GET(mtx, 3, 3)											// ~198 / 200
		);
		if( !isBounded_Vector3f( &light_pos_ndc, &(Vector3f) { 1.0f, 1.0f, 1.0f } ) ) {
			continue;
		}

		// position in device coordinates
		GfxMode const* gfxMode = getCurrentMode_GfxDevice( getGfxDevice_Sys() );
		Vector2f const light_pos_scr = create_Vector2f(
			+0.5f * getWidth_Viewport( getViewport_Engine( self ) ) * (light_pos_ndc.x + 1.0f) * gfxMode->width,
			-0.5f * getHeight_Viewport( getViewport_Engine( self ) ) * (light_pos_ndc.y - 1.0f) * gfxMode->height
		);

		// check if occluded
		{
			Surface const* zbuf = getZBuffer_Engine( self );
			if( !coordInBounds_Surface( zbuf, light_pos_scr.x, light_pos_scr.y ) ) { 
				continue;
			}
			uint16_t const zdev = *((uint16_t*)getValue_Surface( zbuf, light_pos_scr.x, light_pos_scr.y ));
			float const zval = zAsFloat_GfxDevice( getGfxDevice_Sys(), zdev );
			if( light_pos_ndc.z >= zval ) {
				continue;
			}
		}

		calculateFlarePositions( flareXf, &flare_nodes, &root_node, &light_pos_ndc );
	}
	destroy_VecIterator( &flares_iterator );

	renderFlares( self, root_node );

	{
		destroy_Vec( &flare_nodes );

		destroy_Vec( &flaresXf );
		destroy_Vec( &tn_list );

		freeRaw_Engine( self, &root );
	}
}

void clearColorbufF_Engine( Engine* self, void* arg ) {
	Vector3f const* color = arg;
	pushSetClearColor_DList( &self->dl,
		color->x,
		color->y,
		color->z,
		1.0f
	);
	pushClearBuffers_DList( &self->dl, Color_BufferType_DList );
}

void clearZbufF_Engine( Engine* self, void* arg ) {
	float const* zval = arg;
	pushSetClearZValue_DList( &self->dl, *zval );
	pushClearBuffers_DList( &self->dl, Z_BufferType_DList );
}

void processTask_Engine( Engine* self, Task const* task ) {
	reset_DList( &self->dl );
	pushStart_DList( &self->dl );
	pushSetViewport_DList( &self->dl, getViewport_Engine( self ) );

	{
		Surface const* colorSurface = getRenderTarget_Engine( self );
		pushSetBuffer_DList(
			&self->dl,
			Color_BufferType_DList,
			0,
			getWidth_Surface( colorSurface ),
			getHeight_Surface( colorSurface ),
			getData_Surface( colorSurface )
		);
	}

	{
		Surface const* zSurface = getZBuffer_Engine( self );
		pushSetBuffer_DList(
			&self->dl,
			Z_BufferType_DList,
			0,
			zSurface ? getWidth_Surface( zSurface ) : 0,
			zSurface ? getHeight_Surface( zSurface ) : 0,
			zSurface ? getData_Surface( zSurface ) : NULL
		);
	}

	task->f( (struct Engine*)self, task->arg );

	pushSync_DList( &self->dl );
	pushEnd_DList( &self->dl );
}

Xformer* createXformer_Engine( Engine* self, ProcessFunc_Xformer processor ) {
	Xformer* xf = allocate_MemPool( &self->xformersPool );
	init_Xformer( xf, processor );

	return xf;
}

void freeXformer_Engine( Engine* self, Xformer* xf ) {
	destroy_Xformer( xf );
	return_MemPool( &self->xformersPool, xf );
}

Raw_Xformer createRaw_Engine( Engine* self ) {
	return (Raw_Xformer) {
		.tag = Raw_ObjectType,
		.xf = createXformer_Engine( self, NULL )
	};
}

void freeRaw_Engine( Engine* self, Raw_Xformer* rawXf ) {
	freeXformer_Engine( self, rawXf->xf );
}

// TODO: consider move semantics for staticMesh?
// BUT: StaticMesh own a lot of data, which should be shared (geometry, textures).
StaticMesh_Xformer createStaticMesh_Engine( Engine* self, StaticMesh const* staticMesh_ ) {
	StaticMesh* staticMesh = allocate_MemPool( &self->staticMeshesPool );
	*staticMesh = *staticMesh_;

	return (StaticMesh_Xformer) {
		.tag = StaticMesh_ObjectType,
		.xf = createXformer_Engine( self, NULL ),
		.staticMesh = staticMesh
	};
}

void freeStaticMesh_Engine( Engine* self, StaticMesh_Xformer* staticMeshXf ) {
	freeXformer_Engine( self, staticMeshXf->xf );
	return_MemPool( &self->staticMeshesPool, staticMeshXf->staticMesh );
}

Mesh_Xformer createMesh_Engine( Engine* self,
	Mesh				meshFrom,
	Material const*		material,
	VertexShader const*	vertexShader
) {
	Mesh* mesh = allocate_MemPool( &self->meshesPool );
	*mesh = meshFrom;

	if( material ) {
		setMaterialId_Layer( getLayer_Mesh( mesh, 0 ),
			addMaterialReference_Engine( self, material )
		);
	}

	return (Mesh_Xformer) {
		.tag = Mesh_ObjectType,
		.xf = createXformer_Engine( self, NULL ),
		.mesh = mesh,
		.vertexShader = vertexShader
	};
}

void freeMesh_Engine( Engine* self, Mesh_Xformer* meshXf ) {
	freeXformer_Engine( self, meshXf->xf );
	destroy_Mesh( meshXf->mesh );
	return_MemPool( &self->meshesPool, meshXf->mesh );
}

Billboard_Xformer createBillboard_Engine( Engine* self, StaticMesh const* staticMesh ) {
	Billboard* bb = allocate_MemPool( &self->billboardsPool );
	init_Billboard( bb, staticMesh );

	return (Billboard_Xformer) {
		.tag = Billboard_ObjectType,
		.xf = createXformer_Engine( self, processBillboard_Xformer ),
		.billboard = bb,
	};
}

Camera_Xformer createCamera_Engine( Engine* self, CameraType type ) {
	Camera* camera = allocate_MemPool( &self->camerasPool );
	init_Camera( camera, type );

	return (Camera_Xformer) {
		.tag = Camera_ObjectType,
		.xf = createXformer_Engine( self, processCamera_Xformer ),
		.camera = camera
	};
}

void freeCamera_Engine( Engine* self, Camera_Xformer* cameraXf ) {
	destroy_Camera( cameraXf->camera );
	return_MemPool( &self->camerasPool, cameraXf->camera );
}

Lux_Xformer createLux_Engine( Engine* self ) {
	Lux* lux = allocate_MemPool( &self->luxesPool );
	init_Lux( lux );

	return (Lux_Xformer) {
		.tag = Lux_ObjectType,
		.xf = createXformer_Engine( self, NULL ),
		.lux = lux,
	};
}

void freeLux_Engine( Engine* self, Lux_Xformer* luxXf ) {
	destroy_Lux( luxXf->lux );
	return_MemPool( &self->luxesPool, luxXf->lux );
}

Flare_Xformer createFlare_Engine( Engine* self, Vec* flareMeshes ) {
	return (Flare_Xformer) {
		.tag = Flare_ObjectType,
		.xf = createXformer_Engine( self, NULL ),
		.meshes = flareMeshes
	};
}

Skybox_Xformer createSkybox_Engine( Engine* self, Mesh meshFrom, Material const* material ) {
	Mesh* mesh = allocate_MemPool( &self->meshesPool );
	*mesh = meshFrom;

	setMaterialId_Layer( (Layer*)getHead_Vec( getLayers_Mesh( mesh ) ),
		material ? addMaterialReference_Engine( self, material ) : 0
	);

	Skybox* skybox = (Skybox*)malloc( sizeof(*skybox) );
	init_Skybox( skybox, mesh );

	return (Skybox_Xformer) {
		.tag = Skybox_ObjectType,
		.xf = createXformer_Engine( self, processSkybox_Xformer ),
		.skybox = skybox
	};
}

Material* createMaterial_Engine( Engine* self,
	Vector4f const			color,
	TextureCombine const	combineMode,
	uint16_t const			texture0Id,
	uint16_t const			texture1Id
) {
	Material* material = allocate_MemPool( &self->materialsPool );
	init_Material( material, color, combineMode, texture0Id, texture1Id );

	return material;
}
