#include <assert.h>
#include <interpol/interpol.h>
#include <math/math.h>
#include "modeler/modeler.h"

void translateMesh_Modeler( Mesh* mesh, Vector3f const* translation ) {
	VecIterator layers = createLayerIterator_Mesh( mesh );
	for( Layer* layer; (layer = (Layer*)next_VecIterator( &layers )); ) {

		VecIterator verts = createVertexIterator_Layer( layer );
		for( Vertex_generic* vertex; (vertex = (Vertex_generic*)next_VecIterator( &verts )); ) {
			vertex->pos = add_Vector3f( &vertex->pos, translation );
		}
		destroy_VecIterator( &verts );
	}
	destroy_VecIterator( &layers );
}

void scaleMesh_Modeler( Mesh* mesh, Vector3f const* scale ) {
	VecIterator layers = createLayerIterator_Mesh( mesh );
	for( Layer* layer; (layer = (Layer*)next_VecIterator( &layers )); ) {

		VecIterator verts = createVertexIterator_Layer( layer );
		for( Vertex_generic* vertex; (vertex = (Vertex_generic*)next_VecIterator( &verts )); ) {
			vertex->pos = (Vector3f) {
				.x = vertex->pos.x * scale->x,
				.y = vertex->pos.y * scale->y,
				.z = vertex->pos.z * scale->z
			};
		}
		destroy_VecIterator( &verts );
	}
	destroy_VecIterator( &layers );
}

static void appendStrip( Layer* layer, uint_t const baseIdx, uint_t const sidesNum, uint_t const upCorrection ) {
	switch( getPrimitiveType_Layer( layer ) ) {
		case Triangle_PrimitiveType_Layer:
			for( size_t sideIdx = 0; sideIdx < sidesNum; ++sideIdx ) {
				addTriangle_Layer( layer,
					(Triangle) {
						.ids[0] = baseIdx + sideIdx,
						.ids[1] = baseIdx + upCorrection + sidesNum + sideIdx,
						.ids[2] = baseIdx + upCorrection + sidesNum + (sideIdx + 1) % sidesNum
					}
				);
				addTriangle_Layer( layer,
					(Triangle) {
						.ids[0] = baseIdx + sideIdx,
						.ids[1] = baseIdx + upCorrection + sidesNum + (sideIdx + 1) % sidesNum,
						.ids[2] = baseIdx + (sideIdx + 1) % sidesNum
					}
				);
			}
			break;
		case Segment_PrimitiveType_Layer:
			for( size_t sideIdx = 0; sideIdx < sidesNum; ++sideIdx ) {
				addSegment_Layer( layer,
					(Segment) {
						.ids[0] = baseIdx + sideIdx,
						.ids[1] = baseIdx + upCorrection + sidesNum + sideIdx
					}
				);
			}
			break;
	}
}

Mesh createSegment_Modeler( Mesh const* base, Vector3f const* normal ) {
	Mesh result = clone_Mesh( base );
	translateMesh_Modeler( &result, normal );
	assign_Mesh( &result, createUnion_Modeler( base, &result ) );
	appendStrip(
		getLayer_Mesh( &result, 0 ),
		0,
		getVerticesNum_Layer( getLayerConst_Mesh( base, 0 ) ),
		0
	);

	return result;
}

Mesh createTetrahedron_Modeler( void ) {
	Layer layer = create_Layer( Triangle_PrimitiveType_Layer );
	Mesh mesh = create_Mesh();

	{
		Vector3f const v[] = {
			create_Vector3f( 0.0f, 1.0f, 0.0f ),
			create_Vector3f( 0.0f, -0.5f, -sqrtf( 0.75f ) ),
			create_Vector3f( -sqrtf( 0.75f ), -0.5f, 0.5f ),
			create_Vector3f( sqrtf( 0.75f ), -0.5f, 0.5f )
		};
		Vec verts = createFromArray_Vec(
			(Vertex_generic[]) {

				// face 0
				[0] = (Vertex_generic) {
					.pos		= v[0],
					.color		= create_Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= v[0],
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[1] = (Vertex_generic) {
					.pos		= v[2],
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= v[2],
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[2] = (Vertex_generic) {
					.pos		= v[3],
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= v[3],
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},

				// face 1
				[3] = (Vertex_generic) {
					.pos		= v[0],
					.color		= create_Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= v[0],
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[4] = (Vertex_generic) {
					.pos		= v[3],
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= v[3],
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[5] = (Vertex_generic) {
					.pos		= v[1],
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= v[1],
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				// face 2
				[6] = (Vertex_generic) {
					.pos		= v[0],
					.color		= create_Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= v[0],
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[7] = (Vertex_generic) {
					.pos		= v[1],
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= v[1],
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[8] = (Vertex_generic) {
					.pos		= v[2],
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= v[2],
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},

				// face 3
				[9] = (Vertex_generic) {
					.pos		= v[1],
					.color		= create_Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= v[1],
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[10] = (Vertex_generic) {
					.pos		= v[2],
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= v[2],
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[11] = (Vertex_generic) {
					.pos		= v[3],
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= v[3],
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
			},
			sizeof(Vertex_generic), Default_RegionAlignment, 12
		);
		addVertices_Layer( &layer, &verts );
		destroy_Vec( &verts );
	}

	{
		Vec triangles = createFromArray_Vec(
			(Triangle[]) {
				[0] = create_Triangle( 0, 1, 2 ),
				[1] = create_Triangle( 3, 4, 5 ),
				[2] = create_Triangle( 6, 7, 8 ),
				[3] = create_Triangle( 9, 10, 11 ),
			},
			sizeof(Triangle), Default_RegionAlignment, 4
		);
		addTriangles_Layer( &layer, &triangles );
		destroy_Vec( &triangles );
	}

	addLayer_Mesh( &mesh, layer );

	return mesh;
}

Mesh createCube_Modeler( bool const createPolygons, bool const createSegments ) {
	Layer trianglesLayer = create_Layer( Triangle_PrimitiveType_Layer );
	Layer segmentsLayer = create_Layer( Segment_PrimitiveType_Layer );

	{
		Vec verts = createFromArray_Vec(
			(Vertex_generic[]) {
				// front
				[0] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[1] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, -1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[2] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, 1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				[3] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, 1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 1.0f )
				},

				// back
				[4] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, -1.0f, 1.0f ),
					//.color		= create_Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 1.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},
				
				[5] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, -1.0f, 1.0f ),
					//.color		= create_Vector4f( 1.0f, 0.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 1.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[6] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, 1.0f, 1.0f ),
					//.color		= create_Vector4f( 1.0f, 1.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 1.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				[7] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, 1.0f, 1.0f ),
					//.color		= create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 1.0f )
				},

				// left
				[8] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[9] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, -1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[10] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, 1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				[11] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, 1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 1.0f )
				},

				// right
				[12] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, -1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[13] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, -1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[14] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, 1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				[15] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, 1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 1.0f )
				},

				// bottom
				[16] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, -1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[17] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, -1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[18] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, -1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				[19] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 1.0f )
				},

				// top
				[20] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, 1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 0.0f )
				},

				[21] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, 1.0f, -1.0f ),
					//.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 0.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 0.0f )
				},

				[22] = (Vertex_generic) {
					.pos		= create_Vector3f( 1.0f, 1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 0.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 1.0f, 1.0f )
				},
				
				[23] = (Vertex_generic) {
					.pos		= create_Vector3f( -1.0f, 1.0f, 1.0f ),
					//.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.color		= create_Vector4f( 0.0f, 1.0f, 1.0f, 1.0f ),
					.normal		= create_Vector3f( -1.0f, -1.0f, -1.0f ),
					.texcoords	= create_Vector2f( 0.0f, 1.0f )
				},
			},
			sizeof(Vertex_generic), Default_RegionAlignment, 24
		);

		if( true == createPolygons ) {
			addVertices_Layer( &trianglesLayer, &verts );
		}

		if( true == createSegments ) {
			addVertices_Layer( &segmentsLayer, &verts );
		}

		destroy_Vec( &verts );
	}

	Mesh cube = create_Mesh();
	if( true == createPolygons ) {
		Vec triangles = createFromArray_Vec(
			(Triangle[]) {
				[0] = create_Triangle( 0, 1, 2 ),
				[1] = create_Triangle( 0, 2, 3 ),

				[2] = create_Triangle( 4, 6, 5 ),
				[3] = create_Triangle( 4, 7, 6 ),

				[4] = create_Triangle( 8, 10, 9 ),
				[5] = create_Triangle( 8, 11, 10 ),

				[6] = create_Triangle( 12, 13, 14 ),
				[7] = create_Triangle( 12, 14, 15 ),

				[8] = create_Triangle( 16, 17, 18 ),
				[9] = create_Triangle( 16, 18, 19 ),

				[10] = create_Triangle( 20, 21, 22 ),
				[11] = create_Triangle( 20, 22, 23 )
			},
			sizeof(Triangle), Default_RegionAlignment, 12
		);
		addTriangles_Layer( &trianglesLayer, &triangles );
		destroy_Vec( &triangles );

		addLayer_Mesh( &cube, trianglesLayer );
	}

	if( true == createSegments ) {
		Vec segments = createFromArray_Vec(
			(Segment[]) {
				[0] = create_Segment( 0, 1 ),
				[1] = create_Segment( 1, 2 ),
				[2] = create_Segment( 2, 3 ),
				[3] = create_Segment( 3, 0 ),

				[4] = create_Segment( 4, 5 ),
				[5] = create_Segment( 5, 6 ),
				[6] = create_Segment( 6, 7 ),
				[7] = create_Segment( 7, 4 ),

				[8] = create_Segment( 8, 9 ),
				[9] = create_Segment( 10, 11 ),

				[10] = create_Segment( 12, 13 ),
				[11] = create_Segment( 14, 15 ),
			},
			sizeof(Segment), Default_RegionAlignment, 12
		);
		addSegments_Layer( &segmentsLayer, &segments );
		destroy_Vec( &segments );

		addLayer_Mesh( &cube, segmentsLayer );
	}

	return cube;
}

Mesh createSphere_Modeler( size_t const xDiv, size_t const zDiv ) {
	Layer layer = create_Layer( Triangle_PrimitiveType_Layer );
	Mesh sphere = create_Mesh();
	addLayer_Mesh( &sphere, layer );

	{
		float const delta_x = M_PI / (float)zDiv;
		float angle_x = 0.0f;
		for( size_t z=0; z<=zDiv; ++z, angle_x += delta_x ) {
			Mesh circle = createCircle_Modeler(
				fabsf( sinf( angle_x ) ),
				xDiv,
				true,
				true, !true
			);
			for( size_t vertexIdx=0; vertexIdx<xDiv; ++vertexIdx ) {
				Vertex_generic* vertex = getVertex_Layer( getLayer_Mesh( &circle, 0 ), vertexIdx );
				vertex->color = (Vector4f) {
					.x = 1.0f / (vertexIdx+1),
					.y = 1.0f / (vertexIdx+1),
					.z = 1.0f / (vertexIdx+1),
					.t = 1.0f
				};
			}
			translateMesh_Modeler( &circle,
				&(Vector3f) {
					.x = 0.0f,
					.y = 0.0f,
					.z = cosf( angle_x )
				}
			);

			Mesh tmp = createUnion_Modeler( &sphere, &circle );
			destroy_Mesh( &sphere );
			sphere = tmp;

			destroy_Mesh( &circle );
		}
	}

	{
		Layer* layer = getLayer_Mesh( &sphere, 0 );
		for( size_t segmentIdx=0; segmentIdx<zDiv; ++segmentIdx ) {
			appendStrip( layer, xDiv * segmentIdx, xDiv, 0 );
		}
	}

	return sphere;
}

Mesh createPlane_Modeler( size_t const xDivision, size_t const yDivision ) {
	assert( (0 != xDivision) && (0 != yDivision) );

	Mesh plane = create_Mesh();
	Layer layer = create_Layer( Triangle_PrimitiveType_Layer );

	float dx = 2.0f / xDivision;
	float dy = 2.0f / yDivision;
	for( size_t y=0; y<=yDivision; ++y ) {
		for( size_t x=0; x<=xDivision; ++x ) {
			addVertex_Layer( &layer, (Vertex_generic) {
				.pos = {
					.x = -1.0f + x*dx,
					.y = -1.0f + y*dy, 
					.z = 0.0f
				}, 
				.color = { .x = 1.0f, .y = 1.0f, .z = 1.0f, .t = 1.0f },
				.texcoords = { 
					.x = 0.0f + 0.5f*x*dx,
					.y = 1.0f - 0.5f*y*dy
				}
			});
		}
	}

	for( size_t y=0; y<yDivision; ++y ) {
		//if( (y & 0x1) == 1 ) { continue; }
		for( size_t x=0; x<xDivision; ++x ) {
			addTriangle_Layer( &layer,
				(Triangle const) {
					.ids[0] = ((y+1)*(1 + xDivision) + x + 1),
					.ids[1] = (y*(1 + xDivision) + x + 1),
					.ids[2] = (y*(1 + xDivision) + x)
				}
			);
			addTriangle_Layer( &layer,
				(Triangle const) {
					.ids[0] = ((y+1)*(1 + xDivision) + x),
					.ids[1] = ((y+1)*(1 + xDivision) + x + 1),
					.ids[2] = (y*(1 + xDivision) + x)
				}
			);
		}
	}

	addLayer_Mesh( &plane, layer );

	return plane;
}

Mesh createCircle_Modeler( float const radius, size_t const sidesNum, bool const vertsOnly, bool const createPolygons, bool const createSegments ) {
	assert( createPolygons ^ createSegments );

	Layer layer = create_Layer( createPolygons ? Triangle_PrimitiveType_Layer : Segment_PrimitiveType_Layer );

	float const step = 2*M_PI / sidesNum;
	size_t const vertsNum = sidesNum + (false == vertsOnly);
	float angle = 0.0f;
	for( size_t i = 0; i < vertsNum; ++i, angle += step ) {
		addVertex_Layer( &layer, (Vertex_generic) {
			.pos = (sidesNum == i) ?
				create_Vector3f( 0.0f, 0.0f, 0.0f ) :
				create_Vector3f( radius * cosf( angle ), radius * sinf( angle ), 0.0f ),
			.color = create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f )
		});
	}

	Mesh circle = create_Mesh();
	if( (false == vertsOnly) && createPolygons ) {
		for( size_t i = 0; i < sidesNum; ++i ) {
			addTriangle_Layer( &layer,
				create_Triangle( sidesNum, i, (1+i) % sidesNum )
			);
		}
	} else if( (false == vertsOnly) && createSegments ) {
		for( size_t i = 0; i < sidesNum; ++i ) {
			addSegment_Layer( &layer,
				create_Segment( i, (1+i) % sidesNum )
			);
		}
	}

	addLayer_Mesh( &circle, layer );

	return circle;
}

RingMesh createRing_Modeler( size_t const sidesNum, float const innerRadius, float const outerRadius ) {
	Mesh innerCircle = createCircle_Modeler( innerRadius, sidesNum, true, true, false );
	Mesh outerCircle = createCircle_Modeler( outerRadius, sidesNum, true, true, false );

	Mesh ring = createUnion_Modeler( &innerCircle, &outerCircle );
	Layer* layer = getLayer_Mesh( &ring, 0 );
	for( size_t side = 0; side < sidesNum; ++side ) {
		addTriangle_Layer( layer,
			create_Triangle( side, side + sidesNum, (side == sidesNum - 1) ? sidesNum : (1 + side + sidesNum) )
		);
		addTriangle_Layer( layer,
			create_Triangle( side, (side == sidesNum - 1) ? sidesNum : (1 + side + sidesNum), (1 + side) % sidesNum )
		);
	}

	destroy_Mesh( &outerCircle );
	destroy_Mesh( &innerCircle );

	return create_RingMesh( ring );
}

Mesh createCylinder_Modeler( float const radius, size_t const sidesNum, size_t const segmentsNum, Type_Primitives_Layer const primitiveType ) {
	bool const
		createPolygons = (Triangle_PrimitiveType_Layer == primitiveType),
		createSegments = (Segment_PrimitiveType_Layer == primitiveType);

	Mesh result = createCircle_Modeler( radius, sidesNum, false, createPolygons, createSegments );			// base
	{
		Mesh circle = createCircle_Modeler( radius, sidesNum, !false, createPolygons, createSegments );

		for( size_t segmentIdx = 0, baseIdx = 0; segmentIdx < segmentsNum; ++segmentIdx ) {
			translateMesh_Modeler( &circle, &(Vector3f) { 0.0f, 0.0f, 1.0f } );

			assign_Mesh( &result, createUnion_Modeler( &result, &circle ) );

			uint_t const upCorrection = (0 == segmentIdx) ? 1 : 0;
			appendStrip(
				getLayer_Mesh( &result, 0 ),
				baseIdx,
				sidesNum,
				upCorrection
			);

			baseIdx += sidesNum + upCorrection;
		}

		destroy_Mesh( &circle );
	}

	{
		Mesh circle = createCircle_Modeler( radius, sidesNum, false, createPolygons, createSegments );

		translateMesh_Modeler( &circle, &(Vector3f) { 0.0f, 0.0f, (float)segmentsNum } );
		assign_Mesh( &result, createUnion_Modeler( &result, &circle ) );

		destroy_Mesh( &circle );
	}

	return result;
}

Mesh createArrow_Modeler( void ) {
	Mesh plane = createPlane_Modeler( 1, 1 );
	scaleMesh_Modeler( &plane, &(Vector3f) { .x = 0.75f, .y = 0.4f, .z = 1.0f } );
	translateMesh_Modeler( &plane, &(Vector3f) { .x = -0.75f, .y = 0.0f, .z = 0.0f } );

	Mesh triangle = createCircle_Modeler( 1.0f, 3, false, true, false );
	translateMesh_Modeler( &triangle, &(Vector3f) { .x = 0.5f, .y = 0.0f, .z = 0.0f } );

	return createUnion_Modeler( &plane, &triangle );
}

Mesh createUnion_Modeler( Mesh const* meshA, Mesh const* meshB ) {
	assert(
		getLayersNum_Mesh( meshA ) == 1 &&
		getLayersNum_Mesh( meshB ) == 1
	)

	Mesh result = clone_Mesh( meshA );
	Layer* layerResult = getLayer_Mesh( &result, 0 );
	Layer const* layerB = getLayerConst_Mesh( meshB, 0 );

	if( getMaterialId_Layer( layerResult ) != getMaterialId_Layer( layerB ) ) {
		addLayer_Mesh( &result, clone_Layer( layerB ) );
		return result;
	}

	size_t const vtxIdxOffset = getVerticesNum_Layer( layerResult );
	addVertices_Layer( layerResult,
		getVerticesConst_Layer( layerB )
	);

	if( Triangle_PrimitiveType_Layer == getPrimitiveType_Layer( layerResult ) ) {
		size_t const triIdxBase = getTrianglesNum_Layer( layerResult );
		addTriangles_Layer( layerResult,
			getTrianglesConst_Layer( layerB )
		);

		VecIterator triangles = createTriangleIterator_Layer( layerResult,
			(IteratorRange) { .type = From_IteratorRange, .from = triIdxBase }
		);
		for( Triangle* triangle; (triangle = next_VecIterator( &triangles )); ) {
			triangle->ids[0] += vtxIdxOffset;
			triangle->ids[1] += vtxIdxOffset;
			triangle->ids[2] += vtxIdxOffset;
		}
		destroy_VecIterator( &triangles );
	}

	if( Segment_PrimitiveType_Layer == getPrimitiveType_Layer( layerResult ) ) {
		size_t const segIdxBase = getSegmentsNum_Layer( layerResult );
		addSegments_Layer( layerResult,
			getSegmentsConst_Layer( layerB )
		);

		VecIterator segments = createSegmentIterator_Layer( layerResult,
			(IteratorRange) { .type = From_IteratorRange, .from = segIdxBase }
		);
		for( Segment* segment; (segment = next_VecIterator( &segments )); ) {
			segment->ids[0] += vtxIdxOffset;
			segment->ids[1] += vtxIdxOffset;
		}
		destroy_VecIterator( &segments );
	}

	return result;
}

Vertex_generic subdivideSegment_Modeler( Vertex_generic const* v0, Vertex_generic const* v1 ) {
	return (Vertex_generic) {
		.pos		= calculate_LERP3f( &v0->pos, &v1->pos, 0.5f ),
		.color		= calculate_LERP4f( &v0->color, &v1->color, 0.5f ),
		.normal		= create_Vector3f ( 1.0f, 1.0f, 1.0f ),
		.texcoords	= calculate_LERP2f( &v0->texcoords, &v1->texcoords, 0.5f )
	};
}

typedef struct {
	Vertex_generic verts[3];
} Vertex_generic_Triple;

static Vertex_generic_Triple subdivideTriangle_Modeler( Vertex_generic const* v0, Vertex_generic const* v1, Vertex_generic const* v2 ) {
	return (Vertex_generic_Triple) {
		.verts[0] = subdivideSegment_Modeler( v0, v1 ),
		.verts[1] = subdivideSegment_Modeler( v1, v2 ),
		.verts[2] = subdivideSegment_Modeler( v2, v0 )
	};
}

void subdivideTriangle_Layer( Layer* self, size_t const triangleIndex ) {
	Triangle* triangle = getTriangle_Layer( self, triangleIndex );

	Vec newVerts = createFromArray_Vec(
		subdivideTriangle_Modeler(
			getVertexConst_Layer( self, triangle->ids[0] ),
			getVertexConst_Layer( self, triangle->ids[1] ),
			getVertexConst_Layer( self, triangle->ids[2] )
		).verts,
		sizeof(Vertex_generic), Default_RegionAlignment, 3
	);
	size_t const indexOfAdded = addVertices_Layer( self, &newVerts );
	destroy_Vec( &newVerts );

	Vec newTriangles = createFromArray_Vec(
		(Triangle[]) {
			create_Triangle( triangle->ids[0], 0 + indexOfAdded, 2 + indexOfAdded ),
			create_Triangle( triangle->ids[1], 1 + indexOfAdded, 0 + indexOfAdded ),
			create_Triangle( triangle->ids[2], 2 + indexOfAdded, 1 + indexOfAdded )
		},
		sizeof(Triangle), Default_RegionAlignment, 3
	);
	addTriangles_Layer( self, &newTriangles );
	destroy_Vec( &newTriangles );
	assign_Triangle( triangle,
		create_Triangle( 0 + indexOfAdded, 1 + indexOfAdded, 2 + indexOfAdded )
	);
}

void subdivideMesh_Modeler( Mesh* self ) {
	Layer* layer = getLayer_Mesh( self, 0 );
	subdivideTriangle_Layer( layer, 0 );
}

//void extrude_Modeler( Mesh* mesh, Set const* face_indices ) {
//}
