#include <assert.h>
#include "string.h"
#include "malloc.h"

extern uint32_t __heap_start;
extern uint32_t __heap_end;

typedef struct Header {
	struct Header*	next;
	struct Header*	prev;
	size_t			block_size;
	uint32_t		pad;		// we need this to keep addresses aligned
} Header;

static void* skip_header( Header const* header ) {
	return (void*)((uintptr_t)header + sizeof(*header));
}

static size_t space_between( Header const* header, Header const* next ) {
	return ( NULL == next ) ?
		0 :
		(uintptr_t)next - ( (uintptr_t)skip_header( header ) + (uintptr_t)header->block_size );
}

static size_t align_16( size_t const size ) {
	return (size + 0xF) & (~0xF);
}

static bool is_aligned_16( void const* ptr ) {
	return 0 == ((uintptr_t)ptr & 0xF);
}

// WARNING: this is not thread safe
void* malloc( size_t size ) {
	if( 0 == size ) { return NULL; }

	size = align_16( size );

	static Header* head;
	static Header* tail;
	static bool initialized = false;
	if( !initialized ) {
		assert( 16 == sizeof(Header) );

		head = (Header*)&__heap_start;
		tail = (Header*)&__heap_end;
	
		head->next = tail;
		head->prev = NULL;
		head->block_size = 0;

		tail->next = NULL;
		tail->prev = head;
		tail->block_size = 0;

		initialized = true;
	}

	for( Header* current = head; NULL != current; current = current->next ) {
		if( space_between( current, current->next ) < size + sizeof(Header) ) {
			continue;
		}

		Header* new = (Header*)( (uintptr_t)skip_header( current ) + current->block_size );
		assert( is_aligned_16( new ) );

		new->prev = current;
		new->next = current->next;
		new->block_size = size;

		current->next = new;
		new->next->prev = new;

		return skip_header( new );
	}

	abort();
	return NULL;
}

void free( void* ptr ) {
	assert( is_aligned_16( ptr ) );

	if( NULL == ptr ) { return; }

	Header* header = (Header*)( (uintptr_t)ptr - sizeof(Header) );
	(header->prev)->next = header->next;
	(header->next)->prev = header->prev;
}

void* realloc( void* old_ptr, size_t const size ) {
	if( NULL == old_ptr ) {
		return NULL;
	}
	
	size_t old_size = ((Header*)( (uintptr_t)old_ptr - sizeof(Header) ))->block_size;

	void* new_ptr = malloc( size );
	assert( NULL != new_ptr );

	memcpy( new_ptr, old_ptr, old_size );
	free( old_ptr );

	return new_ptr;
}

