#include <assert.h>
#include <sys/audiodevice.h>
#include <sys/region.h>
#include <ultra64.h>

#define AI_THREAD_ID		6
#define AI_THREAD_PRIORITY	13

static void aiThread( void* arg );

void init_AudioDevice( AudioDevice* self, uint32_t const playbackRate, void* outputBuffer0, void* outputBuffer1, size_t const outputBufferSize ) {
	assert( isAddressAligned( outputBuffer0, B8_RegionAlignment ) );
	assert( isAddressAligned( outputBuffer1, B8_RegionAlignment ) );
	assert( 0 == (outputBufferSize % 8) );
	
	{
		setSource_AudioDevice( self, NULL );
		setPlaybackRate_AudioDevice( self, playbackRate );
		self->outputBufferSize = outputBufferSize;
		self->outputBuffer[0] = outputBuffer0;
		self->outputBuffer[1] = outputBuffer1;	//(void*)((uintptr_t)outputBuffer + outputBufferSize);
		self->currentOutputBufferIndex = 1;
	}

	{
		osCreateThread(
			&self->aiThread,
			AI_THREAD_ID,
			aiThread,
			(void*)self,
			&self->aiThreadStack[AI_THREAD_STACK_SIZE/sizeof(u64)],
			AI_THREAD_PRIORITY
		);
		osStartThread( &self->aiThread );
	}
}

uint32_t getPlaybackRate_AudioDevice( AudioDevice const* self ) {
	return self->playbackRate;
}

void setPlaybackRate_AudioDevice( AudioDevice* self, uint32_t const rate ) {
	self->playbackRate = rate;
	osAiSetFrequency( rate );
}

void const* getSource_AudioDevice( AudioDevice const* self ) {
	return self->sourceCurrent;
}

void setSource_AudioDevice( AudioDevice* self, void const* source ) {
	self->sourceStart = source;
	self->sourceCurrent = source;
}

static void* getBufferToWrite_AudioDevice( AudioDevice const* self ) {
	return self->outputBuffer[ self->currentOutputBufferIndex ];
}

static void* getBufferToConsume_AudioDevice( AudioDevice const* self ) {
	return self->outputBuffer[ 1 ^ self->currentOutputBufferIndex ];
}

static void swapBuffers_AudioDevice( AudioDevice* self ) {
	self->currentOutputBufferIndex ^= 1;
}

static void copySourceToBuffer_AudioDevice( AudioDevice const* self ) {
	uint8_t const* source = getSource_AudioDevice( self );
	if( NULL == source ) { return; }
	uint8_t* destination = (uint8_t*)getBufferToWrite_AudioDevice( self );

	for(
		size_t byteIndex = 0;
		byteIndex < self->outputBufferSize;
		byteIndex += 2, source += 2
	) {
		*destination++ = *(source + 1);
		*destination++ = *(source);
	}
}

static void advanceSource_AudioDevice( AudioDevice* self ) {
	self->sourceCurrent = (void const*)((uintptr_t)self->sourceCurrent + self->outputBufferSize);
}

static void resetSource_AudioDevice( AudioDevice* self ) {
	self->sourceCurrent = self->sourceStart;
}

static void aiThread( void* arg ) {
	AudioDevice* self = arg;

	osCreateMesgQueue( &self->aiMessageQ, &self->aiMessages, 1 );
	osSetEventMesg( OS_EVENT_AI, &self->aiMessageQ, NULL );

	while( true ) {
		//while((osAiGetStatus() & AI_STATUS_DMA_BUSY)) {}	// seems like PJ64 can't handle it properly
		if( getSource_AudioDevice( self ) ) {
			copySourceToBuffer_AudioDevice( self );
			advanceSource_AudioDevice( self );
		} else {
			osYieldThread();
		}
		swapBuffers_AudioDevice( self );
		osWritebackDCache( getBufferToConsume_AudioDevice( self ), self->outputBufferSize );
		s32 const aiStatus = osAiSetNextBuffer(
			(uint8_t*)getBufferToConsume_AudioDevice( self ),
			self->outputBufferSize
		);
		if( (s32)AI_STATUS_FIFO_FULL == aiStatus ) {
			//abort();
		} else {
			/*
			copySourceToBuffer_AudioDevice( self );
			advanceSource_AudioDevice( self );
			swapBuffers_AudioDevice( self );
			osWritebackDCache( getBufferToConsume_AudioDevice( self ), self->outputBufferSize );
			osAiSetNextBuffer(
				(uint8_t*)getBufferToWrite_AudioDevice( self ),
				self->outputBufferSize
			);
			*/
		}

		/* wait till buffer is consumed */
		osRecvMesg( &self->aiMessageQ, NULL, OS_MESG_BLOCK );

		if( (uintptr_t)getSource_AudioDevice( self ) >= (uintptr_t)( self->sourceStart ) + self->outputBufferSize*2*14 ) {
			resetSource_AudioDevice( self );
		}
	}
}

