#include <abort.h>
#include <malloc.h>
#include <string.h>
#include <sys/sys.h>
#include <ultra64.h>

#define STACKSIZE		0x2000
#define	NUM_PI_MSGS		8
#define	MAINTHREAD_ID	3
#define	MAINTHREAD_PRI	10
#define	VBITHREAD_ID	4
#define	VBITHREAD_PRI	12
#define	SITHREAD_ID		5
#define	SITHREAD_PRI	11

#define AUDIOBUFFER_BYTES_PER_SAMPLE	2
#define AUDIOBUFFER_SAMPLES_CHNL		(735*2)
#define AUDIOBUFFER_SIZE				(AUDIOBUFFER_SAMPLES_CHNL * AUDIOBUFFER_BYTES_PER_SAMPLE * 2)

int err_color;

static void idle( void* );

static struct SysState {
	/* thread stuff */
	OSThread	idleThread;
	OSThread	mainThread;
	OSThread	vbiThread;
	OSThread	siThread;
	u64			idleThreadStack[STACKSIZE/sizeof(u64)];
	u64			mainThreadStack[STACKSIZE/sizeof(u64)];
	u64			vbiThreadStack[STACKSIZE/sizeof(u64)];
	u64			siThreadStack[STACKSIZE/sizeof(u64)];

	/* various */
	OSPiHandle*	piHandle;
	//int			frameCount;

	/* messages and message queues */
	OSMesgQueue	siMessageQ;
	OSMesgQueue	piMessageQ;
	OSMesgQueue	dmaMessageQ;
	OSMesg		siMessages;
	OSMesg		piMessages[NUM_PI_MSGS];
	OSMesg		dmaMessages;
	OSIoMesg	dmaIOMessageBuf;

	GfxDevice	gfxDevice;
	AudioDevice	audioDevice;

	/* handlers */
	handler_f	handlerVbi;
	void*		handlerVbiArg;
	handler_f	handlerController;
	void*		handlerControllerArg;
} sysState;

void boot( void ) {
	{
		extern char __bss_start[], __bss_end[];
		assert(
			(0 == ((uintptr_t)__bss_start & 0x3)) &&
			(0 == ((uintptr_t)__bss_end & 0x3))
		);
		for(
			uint32_t* bss = (uint32_t*)__bss_start;
			bss < (uint32_t*)__bss_end;
			++bss
		) {
			*bss = 0;
		}
	}

	osInitialize();
	sysState.piHandle = osCartRomInit();
	osCreateThread(
		&sysState.idleThread,
		1,
		idle,
		(void*)0,
		sysState.idleThreadStack + STACKSIZE/sizeof(u64),
		10
	);
	osStartThread( &sysState.idleThread );
}

static void mainproc_wrapper( void* arg ) {
	extern void mainproc( void* );

	init_Sys();

	mainproc( arg );

	raiseError_Sys();
}

static void siThread( void* arg ) {
	if(arg) {}

	OSContPad rdata[ MAXCONTROLLERS ];

	while( true ) {
		osRecvMesg(
			&sysState.siMessageQ,
			NULL,
			OS_MESG_BLOCK
		);

		osContGetReadData( rdata );

		if( sysState.handlerController ) {
			sysState.handlerController( &rdata[0], sysState.handlerControllerArg );
		}
	}
}

static void vbiThread( void* arg ) {
	if(arg) {}


	/* wait for vertical blank */
	while( true ) {
		osContStartReadData( &sysState.siMessageQ );
		waitForVBI_GfxDevice( &sysState.gfxDevice );
		// osJamMesg() call moved to gfxdevice.c
		//++sysState.frameCount;

		/* handle user */
		if( sysState.handlerVbi ) {
			sysState.handlerVbi( NULL, sysState.handlerVbiArg );
		}
	}
}

static void idle( void* arg ) {
	if(arg) {}

	osCreatePiManager(
		OS_PRIORITY_PIMGR,
		&sysState.piMessageQ,
		sysState.piMessages,
		NUM_PI_MSGS
	);

	osCreateThread(
		&sysState.mainThread,
		MAINTHREAD_ID,
		mainproc_wrapper,
		(void*)0,
		sysState.mainThreadStack + STACKSIZE/sizeof(u64),
		MAINTHREAD_PRI
	);
	osStartThread( &sysState.mainThread );

	osSetThreadPri( 0, OS_PRIORITY_IDLE );

	while( true ) {}
}

int init_Sys( void ) {
	err_color = 0xFF;
	//sysState.frameCount = 0;

	init_GfxDevice( &sysState.gfxDevice );

	{
		void* audioBuffer0 = (void*)((uintptr_t)getMemoryEnd_GfxDevice( &sysState.gfxDevice ) + 1);
		void* audioBuffer1 = (void*)((uintptr_t)audioBuffer0 + AUDIOBUFFER_SIZE);
		init_AudioDevice( &sysState.audioDevice,
			44100,
			audioBuffer0,
			audioBuffer1,
			AUDIOBUFFER_SIZE
		);
	}

	osCreateMesgQueue( &sysState.dmaMessageQ, &sysState.dmaMessages, 1 );

	{
		osCreateMesgQueue( &sysState.siMessageQ, &sysState.siMessages, 1 );
		osSetEventMesg( OS_EVENT_SI, &sysState.siMessageQ, NULL );

		u8 pattern;
		OSContStatus sdata[MAXCONTROLLERS];
		osContInit(
			&sysState.siMessageQ,
			&pattern,
			sdata
		);

		sysState.handlerController = NULL;
		sysState.handlerControllerArg = NULL;
		osCreateThread(
			&sysState.siThread,
			SITHREAD_ID,
			siThread,
			(void*)0,
			sysState.siThreadStack + STACKSIZE/sizeof(u64),
			SITHREAD_PRI
		);
		osStartThread( &sysState.siThread );
	}

	{
		sysState.handlerVbi = NULL;
		sysState.handlerVbiArg = NULL;
		osCreateThread(
			&sysState.vbiThread,
			VBITHREAD_ID,
			vbiThread,
			(void*)0,
			sysState.vbiThreadStack + STACKSIZE/sizeof(u64),
			VBITHREAD_PRI
		);
		osStartThread( &sysState.vbiThread );
	}

	return 0;
}

void raiseError_Sys( void ) {
	Surface const* surface = getSurface_GfxDevice( getGfxDevice_Sys(), Display_SurfaceType_GfxDevice );
	for( size_t i=0; i<getPixelsNum_Surface( surface ); ++i) {
		((uint8_t*)getData_Surface( surface ))[i] = err_color;
	}
}

void registerHandler_Sys( HandlerType_Sys const type, handler_f handler, void* userArg ) {
	switch( type ) {
		case Vbi_HandlerType_Sys: {
			sysState.handlerVbi = handler;
			sysState.handlerVbiArg = userArg;
			break;
		}
		case Controller_HandlerType_Sys: {
			sysState.handlerController = handler;
			sysState.handlerControllerArg = userArg;
			break;
		}
		default: {
			break;
		}
	}
}

Time getTime_Sys( void ) {
	Time time;
	time.msec	= sysState.gfxDevice.frameCount * (1000.0f / 60.0f);
	time.sec	= time.msec / 1000;
	time.msec	%= 1000;
	
	return time;
}

void performDma_Sys( void* dest, void const* src, size_t const size ) {
	sysState.dmaIOMessageBuf = (OSIoMesg) {
		.hdr.pri		= OS_MESG_PRI_NORMAL,
		.hdr.retQueue	= &sysState.dmaMessageQ,
		.dramAddr		= dest,
		.devAddr		= (u32)src,
		.size			= size
	};
	osInvalDCache( dest, size );
	osEPiStartDma(
		sysState.piHandle,
		&sysState.dmaIOMessageBuf,
		OS_READ
	);

	OSMesg dummy;
	osRecvMesg( &sysState.dmaMessageQ, &dummy, OS_MESG_BLOCK );
}

GfxDevice* getGfxDevice_Sys( void ) {
	return &sysState.gfxDevice;
}

AudioDevice* getAudioDevice_Sys( void ) {
	return &sysState.audioDevice;
}
/*
void* getHiMem_Sys( void ) {
	extern uint32_t __stack_adr;
	return (void*)(__stack_adr - STACKSIZE);
}
*/
