#include <dsp/signal.h>
#include <interpol/interpol.h>
#include <modeler/modeler.h>
#include <resmgr/resmgr.h>
#include <ScalarField.h>
#include <ScalarFieldPolygonizer.h>
#include <stdlib.h>
#include <sys/sys.h>
#include "scene_action.h"

//#define SKYBOX_OFF
//#define BLOB_OFF
#define CUBES_NUM	4

typedef struct {
	StaticMesh_Xformer	worldMeshXf,
						torusLargeMeshXf,
						torusSmallMeshXf;
	Mesh_Xformer		blobMeshXf,
						cubeXf[CUBES_NUM];
	Skybox_Xformer		skyboxXf;
} Meshes;

typedef struct {
	Lux_Xformer		redLuxXf,
					blueLuxXf;
} Lights;

typedef struct {
	Vec				flares;
	Flare_Xformer	flareXf;
} Flares;

typedef struct {
	ScalarField				scalarField;
	ScalarFieldPolygonizer	polygonizer;

	Raw_Xformer				rootXf;
	Billboard_Xformer		billboardXf;
	Camera_Xformer			cameraXf;
	Meshes					meshes;
	Lights					lights;
	Flares					flares;
	Tree					sceneTree;

	Signal					sinus;
	SamplerLinear			sampler;
} Context;

static float sinf_wrapped( float t, void const* arg ) {
	if( arg ) {}
	return sinf( t );
}

static float calculate_blobfield( Vector3f const* coords, void const* arg ) {
	float const
		offset = *((float const*)arg),
		x0_squared = (coords->x - offset) * (coords->x - offset),
		x1_squared = (coords->x + offset) * (coords->x + offset),
		y_squared = coords->y * coords->y,
		z_squared = coords->z * coords->z;

	return 0.5f * (
		1.0f / sqrtf( x0_squared + (y_squared + z_squared) ) +
		1.0f / sqrtf( x1_squared + (y_squared + z_squared) )
	);
}

static Meshes init_meshes( Engine* engine, Material* materials[static 1] ) {
	Meshes retval;

	{
		retval.worldMeshXf =
			createStaticMesh_Engine( engine, (StaticMesh*)getResource_ResMgr( "world" ) );
	}

	{
		retval.torusLargeMeshXf =
			createStaticMesh_Engine( engine, (StaticMesh*)getResource_ResMgr( "torus" ) );
		retval.torusLargeMeshXf
			.staticMesh->opacity = 0.5f;
	}

	{
		retval.torusSmallMeshXf =
			createStaticMesh_Engine( engine, (StaticMesh*)getResource_ResMgr( "torus" ) );
	}

	#ifndef BLOB_OFF
	{
		retval.blobMeshXf =
			createMesh_Engine( engine, createArrow_Modeler(), materials[0], NULL );		// arrow: temp
	}
	#endif

	#ifndef SKYBOX_OFF
	{
		retval.skyboxXf =
			createSkybox_Engine( engine, createSphere_Modeler( 8, 8 ), materials[0] );
		setOpacity_Mesh( retval.skyboxXf.skybox->mesh, 0.5f );
	}
	#endif
	{
		for( size_t i=0; i<CUBES_NUM; ++i ) {
			retval.cubeXf[i] =
				createMesh_Engine( engine, createCube_Modeler( false, true ), materials[1], NULL );
			if( 2 == i ) { setOpacity_Mesh( retval.cubeXf[i].mesh, 0.6f ); }
		}
	}

	return retval;
}

static Lights init_lights( Engine* engine ) {
	Lights retval;

	{
		retval.redLuxXf = createLux_Engine( engine );
		setColor_Lux( retval.redLuxXf.lux, 0.9f, 0.0f, 0.0f );
	}

	{
		retval.blueLuxXf = createLux_Engine( engine );
		setColor_Lux( retval.blueLuxXf.lux, 0.0f, 0.0f, 0.9f );
	}

	return retval;
}

static Billboard_Xformer init_billboards( Engine* engine ) {
	Billboard_Xformer billboardXf =
		createBillboard_Engine( engine, (StaticMesh*)getResource_ResMgr( "flare1" ) );
	setTextureCombine_Material( &billboardXf.billboard->mesh.materials[0], Alpha_TextureCombine );

	return billboardXf;
}

static void init_flares( Flares* flares, size_t const flareTexturesNum, Engine* engine ) {
	flares->flares = create_Vec(
		sizeof(Pair_StaticMesh_Xformer_Scale),
		Default_RegionAlignment,
		NULL
	);

	for( size_t i=0; i<flareTexturesNum; ++i ) {
		static float const scales[] = { 0.01f, 0.005f, 0.004f, 0.007f, 0.008f, 0.0085f };

		Pair_StaticMesh_Xformer_Scale pair = {
			.staticMeshXf = createStaticMesh_Engine( engine,
				(0 == i) ?
					(StaticMesh const*)getResource_ResMgr( "flare1" ) :
				(1 == i) ?
					(StaticMesh const*)getResource_ResMgr( "flare5" ) :
				(2 == i) ?
					(StaticMesh const*)getResource_ResMgr( "flare6" ) :

					(StaticMesh const*)getResource_ResMgr( "flare5" )
				),
			.scale = scales[i]*1.7f
		};
		setTextureCombine_Material( &pair.staticMeshXf.staticMesh->materials[0], Alpha_TextureCombine );
		pushBack_Vec( &flares->flares, &pair );
	}

	flares->flareXf = createFlare_Engine( engine, &flares->flares );
}

void* init_scene_action( Time const* timeFromStart, Time const* timeDuration, void* arg ) {
	if( timeFromStart ) {}
	if( timeDuration ) {}

	SceneActionArgs const* args = (SceneActionArgs const*)arg;
	Context* context = malloc( sizeof(*context) );

	{
		init_ScalarField( &context->scalarField,
			args->scalar_field_resolution,
			args->scalar_field_resolution,
			args->scalar_field_resolution
		);
		init_ScalarFieldPolygonizer( &context->polygonizer );
	}

	Engine* engine = args->engine;

	{
		context->rootXf = createRaw_Engine( engine );
	}

	{
		context->cameraXf = createCamera_Engine( engine, Perspective_CameraType );
		setParams_PerspCamera( getPersp_Camera( context->cameraXf.camera ),
			63.74f, 1.33f, 1000.0f
		);
	}

	Material* materials[2];
	{
		materials[0] = createMaterial_Engine( engine,
			create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f ),
			Decal_TextureCombine,
			TEXTURE_ID_NONE,
			TEXTURE_ID_NONE
		);
		setLighting_Material( materials[0], false );
		setRenderingSide_Material( materials[0], Cw_RenderingSide );

		materials[1] = createMaterial_Engine( engine,
			create_Vector4f( 1.0f, 1.0f, 1.0f, 1.0f ),
			Decal_TextureCombine,
			addTextureReference_Engine( engine,
				&((StaticMesh*)getResource_ResMgr( "logo" ))->textures[0]
			),
			TEXTURE_ID_NONE
		);
		setLighting_Material( materials[1], false );
		setRenderingSide_Material( materials[1], Cw_RenderingSide );
	}

	context->meshes = init_meshes( engine, materials );
	context->lights = init_lights( engine );
	context->billboardXf = init_billboards( engine );
	init_flares( &context->flares, 6, engine );

	init_Tree( &context->sceneTree,
		create_TreeNode( &context->rootXf, NULL,
			(TreeNode[]) {
				create_TreeNode( &context->cameraXf, NULL, NULL, 0 ),
				create_TreeNode( &context->meshes.worldMeshXf, NULL, NULL, 0 ),
				create_TreeNode( &context->meshes.torusLargeMeshXf, NULL,
					(TreeNode[]) { create_TreeNode( &context->meshes.torusSmallMeshXf, NULL,
							(TreeNode[]) {
								create_TreeNode( &context->billboardXf, NULL, NULL, 0 ), 
								create_TreeNode( &context->lights.redLuxXf, NULL, NULL, 0 ), 
								create_TreeNode( &context->flares.flareXf, NULL, NULL, 0 )
							}, 3
						),
					}, 1
				),
				#ifndef BLOB_OFF
					create_TreeNode( &context->meshes.blobMeshXf, NULL, NULL, 0 ),
				#endif
				#ifndef SKYBOX_OFF
					create_TreeNode( &context->meshes.skyboxXf, NULL, NULL, 0 ),
				#endif
				create_TreeNode( &context->lights.blueLuxXf, NULL, NULL, 0 )
			}, 6
#ifdef BLOB_OFF
			- 1
#endif
#ifdef SKYBOX_OFF
			- 1
#endif
		),
		NULL, NULL
	);
	for( size_t i=0; i<CUBES_NUM; ++i ) {
		attachChild_TreeNode( getRoot_Tree( &context->sceneTree ),
			create_TreeNode( &context->meshes.cubeXf[i], NULL, NULL, 0 )
		);
	}

	{
		context->sinus = create_Signal( sinf_wrapped, NULL );
		context->sampler = create_SamplerLinear(
			create_Sampler( Wrap_SamplerWrapping, 16, 2*M_PI )
		);
	}

	return context;
}

static void handleCamera( Camera_Xformer* cameraXf, SamplerLinear const* sampler, Signal const* sinus, SceneActionArgs const* args, float const time_normalized ) {
	{
		Matrix mv;
		rotate_Matrix( &mv, args->player->angle, 0, 1, 0 );
		transformLocally_Xformer( cameraXf->xf, &mv );
	}

	{
		float const dx = 16.0f * sample_SamplerLinear( sampler, sinus, time_normalized );
		float const dy = 0.0f;

		Matrix mv;
		translate_Matrix( &mv,
			dx + args->player->pos.x,
			dy + args->player->pos.y,
			args->player->pos.z
		);
		transformLocally_Xformer( cameraXf->xf, &mv );
	}
}

static void handle_lights( Lights* lights ) {
	Matrix mv;
	translate_Matrix( &mv, 0, 1, 0 );
	transformLocally_Xformer( lights->redLuxXf.xf, &mv );
}

static void handleTorusMeshes( StaticMesh_Xformer* torusMesh1Xf, StaticMesh_Xformer* torusMesh2Xf, SceneActionArgs const* args, Time const* time ) {
	{
		Matrix mv;
		rotate_Matrix( &mv, *args->angle, 1, 1, 1 );
		transformLocally_Xformer( torusMesh1Xf->xf, &mv );
		translate_Matrix( &mv, 0, 160, 0 );
		transformLocally_Xformer( torusMesh1Xf->xf, &mv );
	}

	{
		Matrix mv;
		scale_Matrix( &mv, &(Vector3f) { 0.4f, 0.4f, 0.4f } );
		transformLocally_Xformer( torusMesh2Xf->xf, &mv );

		Vector3f const pos0 = create_Vector3f( 0.0f, -30.0f, 0.0f );
		Vector3f const pos1 = create_Vector3f( 0.0f, 30.0f, 0.0f );
		Vector3f const pos = calculate_LERP3f( &pos0, &pos1, sinf( 3.2 * time->msec / 1000.f ) );
		translate_Matrix( &mv, pos.x, pos.y, pos.z );
		transformLocally_Xformer( torusMesh2Xf->xf, &mv );
	}
}

static void handleBlobs( Mesh_Xformer* blobmeshXf, ScalarField* scalarField, ScalarFieldPolygonizer const* polygonizer, SceneActionArgs const* args ) {
	float const offset = 0.4*sinf( *args->angle / 100.0f );
	calculate_ScalarField( scalarField, calculate_blobfield, &offset );

	destroy_Mesh( blobmeshXf->mesh );
	*blobmeshXf->mesh =
		process_ScalarFieldPolygonizer( polygonizer, scalarField, 3.08f, 1.0f );

	{
		Matrix mv;
		scale_Matrix( &mv, &(Vector3f) { 200.0f, 200.0f, 200.0f } );
		transformLocally_Xformer( blobmeshXf->xf, &mv );
	}

	{
		Matrix mv;
		translate_Matrix( &mv, 400, 160, 0 );
		transformLocally_Xformer( blobmeshXf->xf, &mv );
	}
}

static void handleSkybox( Skybox_Xformer* skyboxXf, SceneActionArgs const* args ) {
	{
		Matrix mv;
		scale_Matrix( &mv, &(Vector3f) { 900.0f, 900.0f, 900.0f } );
		transformLocally_Xformer( skyboxXf->xf, &mv );
	}

	{
		Matrix mv;
		rotate_Matrix( &mv, *args->angle / 32.0f, 0, 1, 0 );
		transformLocally_Xformer( skyboxXf->xf, &mv );
	}
}

static void handleCubes( size_t numCubes, Mesh_Xformer cubesXf[numCubes], SceneActionArgs const* args ) {
	assert( numCubes <= CUBES_NUM );

	Vector3f const scales[] = {
		create_Vector3f( 40.0f, 40.0f, 40.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 25.0f, 25.0f, 25.0f ),
		create_Vector3f( 30.0f, 30.0f, 30.0f ),
		create_Vector3f( 10.0f, 10.0f, 10.0f ),
		create_Vector3f( 15.0f, 15.0f, 15.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
		create_Vector3f( 20.0f, 20.0f, 20.0f ),
	};
	Vector3f const translations[] = {
		create_Vector3f( 80.0f, 0.0f, 20.0f ),
		create_Vector3f( -160.0f, 10.0f, -40.0f ),
		create_Vector3f( -120.0f, 0.0f, 60.0f ),
		create_Vector3f( -40.0f, 10.0f, 30.0f ),
		create_Vector3f( -180.0f, 0.0f, -20.0f ),
		create_Vector3f( -60.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
		create_Vector3f( -40.0f, 10.0f, 0.0f ),
	};
	Vector3i const axes[] = {
		create_Vector3i( 1, 0, 1 ),
		create_Vector3i( 0, 1, 0 ),
		create_Vector3i( 1, 0, 1 ),
		create_Vector3i( 1, 1, 0 ),
		create_Vector3i( 1, 1, 0 ),
		create_Vector3i( 1, 0, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
		create_Vector3i( 0, 1, 1 ),
	};

	for( size_t i=0; i<numCubes; ++i ) {
		Matrix mv;

		scale_Matrix( &mv, &scales[i] );
		transformLocally_Xformer( cubesXf[i].xf, &mv );

		rotate_Matrix( &mv, (1 + i) * (*args->angle) / 2.0f, axes[i].x, axes[i].y, axes[i].z );
		transformLocally_Xformer( cubesXf[i].xf, &mv );

		translate_Matrix( &mv, translations[i].x, 100.0f + translations[i].y, -100.0f + translations[i].z );
		transformLocally_Xformer( cubesXf[i].xf, &mv );
	}
}

void destroy_scene_action( void* context_generic, void* arg ) {
	if( context_generic ) {}
	if( arg ) {}
	abort2();
}

void handle_scene_action( void* context_generic, Time const* timeFromStart, void* arg ) {

	Context* context = (Context*)context_generic;

	Time const timeDelta = calcDuration_Time(
		&(Time) { .sec = 0, .msec = 1, .nsec = 0 },
		timeFromStart
	);
	float const timeNormalized =
		((float)asMilliseconds_Time( &timeDelta )) /
		asMilliseconds_Time( &(Time) { .sec = 5, .msec = 1, .nsec = 0 }
	);

	SceneActionArgs const* args = (SceneActionArgs const*)arg;
	Engine* engine = args->engine;
	setCamera_Engine( engine, &context->cameraXf );

	handleCamera( &context->cameraXf, &context->sampler, &context->sinus, args, timeNormalized );
	handle_lights( &context->lights );
	handleTorusMeshes( &context->meshes.torusLargeMeshXf, &context->meshes.torusSmallMeshXf, args, timeFromStart );

	#ifndef BLOB_OFF
		handleBlobs( &context->meshes.blobMeshXf, &context->scalarField, &context->polygonizer, args );
	#endif

	#ifndef SKYBOX_OFF
		handleSkybox( &context->meshes.skyboxXf, args );
	#endif

	handleCubes( CUBES_NUM, context->meshes.cubeXf, args );
	
	process_task( engine, &(Task const) {
		.f = (taskfptr)sceneTaskF_Engine,
		.arg = &context->sceneTree
	});
	process_task( engine, &(Task const) {
		.f = (taskfptr)flareTaskF_Engine,
		.arg = &context->sceneTree
	});
}

